package HSPC::Application;

use strict;

use SOAP::Lite;
use DBI;
use HSPC::Logger;
use HSPC::Exc;
use HSPC::MT::GM::Constants qw(SW_GM_DEF_TIMEZONE);

use constant MAX_CACHED_CONNECTIONS => 10;

sub new {
	my ($self, $class, %h) = ({ }, @_);
	if ($ENV{system_obj} && ! $h{reinit_ok}) {
		require HSPC::ProblemReporter;
		my $cs = HSPC::ProblemReporter->_get_stack_trace;
		warn 'Double initialization of ' . __PACKAGE__ . ' package.' .
			" Call stack:\n$cs";
	}
	bless $self, $class;

	## Prepare logging engine
	HSPC::Logger->cleanup() unless $ENV{system_obj};
	###### STORE IN ENV
	$ENV{system_obj} = $ENV{HSPC_SYSTEM} = $self;
}

## Returns stored/newly created SOAP::Lite client object,
## to which ->call method is applicable.
sub get_soap_client {
	SOAP::Lite->proxy('http://127.0.0.1:8080/hspc/xml-api');
}

## DEPRECATED
sub DEPRECATED {
	my $level = (shift || 0) + 1;
	my $sub = shift || (caller $level)[3];
	my $message = $sub . ' is DEPRECATED. ';
	$message .= 'The function is called from ';
	$message .= (caller $level + 1)[3];
	warn $message . "\n";
}

sub trace { DEPRECATED() }

BEGIN {
	no strict 'refs';
	## HSPc code was cleanup using following grep:
	## grep -rP '(\>|\:\:)(trace|log(|_try|_error|_ok|_special|_debug|_info|_warn))([^a-z_]|$)' 
	## To be removed for sp2
	## Create aliases for logging subroutines
	my %aliases = (
		'log'		=> 'info',
		'log_try'	=> 'info',
		'log_error'	=> 'error',
		'log_ok'	=> 'info',
		'log_special'	=> 'info',
		'log_debug'	=> 'debug',
		'log_info'	=> 'info',
		'log_warn'	=> 'warn',
	);

	foreach my $sub (keys %aliases) {
		my $level  = $aliases{$sub};
		my $symbol = __PACKAGE__ . '::' . $sub;
		*{$symbol} = sub {
			DEPRECATED(0, $symbol);
			HSPC::Logger->$level($_[1]);
		}
	}
}

## >> self
## >> message
sub system_die {
	my $self = shift;
	my $message = shift;
	my $t = HSPC::Logger->no_trace;
	$self->generate_bugreport($message);
	throw HSPC::Exc(msg=>$message);
}

## Returns id of last (!) generated bugreport
sub last_bugreport_id { HSPC::Logger->problem_id }

###### Return short report for generate_bugreport() method, can be redefined 
###### in child class
## >> self
## << report
sub _generate_bugreport_common { DEPRECATED() }

## >> self
## >> message
## << problem ID
sub generate_problemreport {
	my $t = HSPC::Logger->no_trace;
	HSPC::Logger->error($_[1]);
	HSPC::Logger->problem_id;
}

## >> self
## >> message
sub generate_bugreport {
	$ENV{layout}->set_error_app if $ENV{layout};
	my $t = HSPC::Logger->no_trace;
	HSPC::Logger->fatal($_[1]);
	HSPC::Logger->problem_id;
}

# =>> mode : /echo/debug/error
# =>  value : 1/0
# <  current mode value
sub text_mode {
	my ($self,%h) = @_;
	$self->{__textout_modes}->{$h{mode}} = $h{value} if $h{mode};
	$self->{__textout_modes}->{$h{mode}};
}

# >  inc : increment offset value to this valye
# <  current mode value
sub text_offset {
	my ($self, $inc) = @_;
	$self->{__textout_offset} += $inc;
	$self->{__textout_offsetc} = '.' x $self->{__textout_offset};
}

# > text
sub text_echo {
	my ($self, $value) = @_;
	return unless $self->{__textout_modes}->{echo};
	print $self->{__textout_offsetc} . "$value\n";
}

# < accounts : all avaliable accounts for current tool
sub user_accounts {
	my ($self, %h) = @_;
	my $all_accounts = $ENV{security_obj}->get_accounts();
	$all_accounts->{ $ENV{security_obj}->account_type };
}

## Returns true, if the database is
## inside a transaction now, false otherwise
## => dbname -- name of the database to check.
sub is_transaction {
	my $self = shift;
	my $dbname = shift;

	return ($ENV{HSPC_DB_TRANS} 
			&& $ENV{HSPC_DB_TRANS}->{$dbname} > 0 
			&& (!$ENV{HSPC_DB_ROLLBACKED}
				|| !$ENV{HSPC_DB_ROLLBACKED}->{$dbname}));
}

## must override
sub is_console { }

# >> class
# => new : open new connection
# < dbh
#  DO NOT CALL $dbh->disconnect() manually ( in open=>0 case )
sub db_connection{
	my ($class,%h) = @_;

	my $dbname = $h{dbname} || 'MAIN';

	$ENV{HSPC_SYSTEM_DBH} = {} unless ref($ENV{HSPC_SYSTEM_DBH});
	my $dbh = $h{new} ? undef : $ENV{HSPC_SYSTEM_DBH}->{$dbname};

	my $old_connection_id = $dbh ? $dbh->{'mysql_thread_id'} : 0;

	require HSPC::Config;
	my $max_countdown = $HSPC::Config::CONFIG{DB_RECONNECT_TIMEOUT} || 10;
	my $countdown = $max_countdown;
	while(!defined($dbh) || !$dbh->FETCH('Active') || !$dbh->ping) {

		## sleep if it's not a first pass.
		if ($countdown < $max_countdown) {
			sleep 1; 
		}

		## disconnect if connection is not working
		if (defined $dbh) {
			eval {
				$countdown == $HSPC::Config::CONFIG{DB_RECONNECT_TIMEOUT}
				? $class->db_disconnect(dbname=>$dbname,force=>1)
				: $dbh->disconnect;
			};
			## die if we are running inside transaction
			if($class->is_transaction($dbname)) {
				$ENV{HSPC_DB_ROLLBACKED}->{$dbname} = 1;
				ref $ENV{HSPC_SYSTEM}
				? $ENV{HSPC_SYSTEM}->system_die("Lost DB connection inside transaction")
				: die "Lost DB connection inside transaction";
			}
		}

		## (re)open connection to DB
		eval { $dbh = $class->connect_once(undef,$dbname) };
		HSPC::Logger->error("Failed to (re)open DB connection: $@")
			if $@ && ref $ENV{HSPC_SYSTEM};
		$countdown--;
		last unless $countdown >= 0;
	}

	$ENV{HSPC_SYSTEM_DBH}->{$dbname} = $dbh unless $h{new};

	my $new_connection_id = $dbh->{'mysql_thread_id'};

	if ($new_connection_id != $old_connection_id){
		$dbh->do("set character set utf8");
		$dbh->do("set names utf8");
	}
	HSPC::Logger->new->debug("SQL Connection got for '$dbname': $old_connection_id => $new_connection_id")
		if $old_connection_id != $new_connection_id;

	return $dbh;
}

## > class
## >>= dbname [opt] - database name to disconnect from
## >>= force  [opt] - force connection to be released
sub db_disconnect {
	my ($class,%h) = @_;
	return unless $ENV{HSPC_SYSTEM_DBH};
	my $sys_dbh;
	if ( $class->is_console() ) {
		$sys_dbh = $ENV{HSPC_SYSTEM_DBH};
	} else {
		## destroy Apache::DBI connection
		$sys_dbh = Apache::DBI->all_handlers();
	}
	## get list of DB to disconnect from
	my @dbnames = $h{dbname} ? ($h{dbname}) : (keys %$sys_dbh);
	@dbnames = ('MAIN') unless scalar @dbnames;

	$ENV{HSPC_DB_TRANS} = {} unless ref($ENV{HSPC_DB_TRANS});
	foreach my $dbname (@dbnames) {
		## disconnect if connected
		if($h{force}) {
			$sys_dbh->{$dbname}->DBI::db::disconnect() if ( $sys_dbh->{$dbname} );
			delete $sys_dbh->{$dbname} unless $class->is_console();
			undef $ENV{HSPC_SYSTEM_DBH}->{$dbname};
		} else {
			$sys_dbh->{$dbname}->disconnect();
		}
		$ENV{HSPC_DB_TRANS}->{$dbname} = 0; ## DB disconnected => all transactions closed
	}
}

# close DB connections 
sub close_all_db_connections {
	if( defined $ENV{HSPC_SYSTEM_DBH}){
		foreach my $dbh (values (%{$ENV{HSPC_SYSTEM_DBH}})) {
			$dbh->disconnect() if defined $dbh;
		}
		$ENV{HSPC_SYSTEM_DBH} = undef;	
	}
}

# close DB connections 
sub force_close_all_db_connections {
	my $self = shift;

	if( defined $ENV{HSPC_SYSTEM_DBH}){
		foreach my $dbh (values (%{$ENV{HSPC_SYSTEM_DBH}})) {
			$dbh->DBI::db::disconnect() if $dbh;
		}
		$ENV{HSPC_SYSTEM_DBH} = undef;	

		if ( !$self->is_console() ) {
			my $sys_dbh = Apache::DBI->all_handlers() || [];
			foreach my $sys_db (keys %$sys_dbh) {
				delete $sys_dbh->{$sys_dbh};
			}
		}
	}
	return undef;
}


sub _template_broken {
	my %args = @_;
	my $ref_arg = $args{arg};
	return "ERROR : template '$$ref_arg' broken at line number $args{lineno}: $args{error}";
}


## >> self
##=>> lang
sub set_current_lang {
	my $self = shift;
	my $lang = shift;

	unless($lang) {
		HSPC::Logger->error("lang is required");
		return;
	}
	$self->{_lang} = $lang;
	$ENV{session}->{LANG} = $self->{_lang} if ( $ENV{session} );
}

sub provider_id { 1 }
sub locale_timezone { SW_GM_DEF_TIMEZONE }

####### store any persistent data
## >> self
## =>> key
## =>> data : scalar or ref to any data
## Returns undef on error
sub store_data {
	my $self = shift;
	my %h = (key => undef, @_);

	unless (defined $h{key}) {
		HSPC::Logger->warn('No key defined!');
		return;
	}

	$ENV{session}->{$h{key}} = $h{data};
}

####### restore any persistent data
## >> self
## =>> key
## < scalar or ref to any data
## returns undef on error
sub restore_data {
	my $self = shift;
	my %h = ( key => undef, @_);

	unless (defined $h{key}) {
		HSPC::Logger->warn('key undefined!');
		return;
	}

	$ENV{session}->{$h{key}};
}

####### connect to DB, set connection encoding _once_
## => dbname -- name of the database, default 'MAIN'
## returns dbh or dies
sub connect_once($$;$) {
	my $self = shift;
	my $r = shift;
	my $dbname = shift || 'MAIN';

	require HSPC::Config;
	my $dbh = DBI->connect(
		'DBI:mysql:' .  $HSPC::Config::CONFIG{$dbname.'_DB_NAME'} .
		';host=' . $HSPC::Config::CONFIG{$dbname.'_DB_HOST'} . 
		';mysql_enable_utf8=1',

		$HSPC::Config::CONFIG{$dbname.'_DB_USER'},
		$HSPC::Config::CONFIG{$dbname.'_DB_PASS'},

		{
			AutoCommit => 1,
			PrintError => 0,
			RaiseError => 1
		},
	);
	die "DBI->connect failed (MySQL server is down?)" unless defined $dbh;
	$dbh->trace(0);
	$dbh;
}

1;

__END__

=pod

=head1 SYNOPSIS

HSPC::Application is a lightweight version of HSPC::System class, intended to be used as CLI toolkit.
In fact, HSPC::System class is derived from it.

=cut
