package HSPC::CompRep;

use strict;

use SWsoft::protect; ## mark this module for encryption upon build

use Data::Dumper;
use Encode;
use File::Copy;
use File::Temp;
use HSPC::Logger;
use Storable qw/dclone/;
use XML::Simple;

## configuration root directory
use constant CR_CONFIG_DIR		=> '/var/opt/hspc-data/Core/CompRep';

## temporary directory
use constant CR_TEMP_DIR		=> '/tmp';

## configuration file extension
use constant CR_CONFIG_EXT		=> '.xml';

## main configuration file name
use constant CR_CONFIG_DEF		=> '_';

## nesting of elements
use constant CR_CONFIG_NESTING	=> [qw/
	root
		director
			manager
				screen
					screen
/];

## format of attributes (either regexp or a value of ref)
use constant CR_NODE_DESCRIPTOR	=> {
	node_id		=> qr/^\d+$/,
	level		=> qr/^\d+$/,
	title_id	=> qr/^\w+$/,
	title_hash	=> 'HASH', ## parameter substitutions for 'title_id'
	uri_params	=> 'HASH', ## additional URI params to use in 'link'
	alias		=> qr/^[\w\-]+$/,
	id			=> qr/^\d\d+$/,
	uri			=> qr/^\d\d(\.\d\d+(\.\d\d+)?)?$/,
	class		=> qr/^\w+(::\w+)*$/,
	method		=> qr/^[\w\d]+$/,
	filter		=> qr/^[\w\d]+$/,
	accounts	=> qr/^\w+(\s*,\s*\w+)*$/,
	groups		=> qr/^\w+(\s*,\s*\w+)*$/,
	icon		=> qr/^[\w\d]+$/,
	helper		=> qr/^[\w\d\.\-]+$/,
	ssl			=> qr/^(on|off)$/,
	referrers	=> qr/^\w+(\s*,\s*\w+)*$/,
};

## list of attributes to inherit values from parent
use constant CR_ATTR_GET_PARENT	=> qw(title_id class method icon accounts groups helper);

## list of attributes to inherit values from 'alias'
use constant CR_ATTR_GET_ALIAS	=> qw(title_id icon);

## list of required attributes
use constant CR_ATTR_REQUIRED	=> qw(alias class method icon helper);

## common prefix of all links
use constant CR_LINK_LOCATION	=> '/hspc';

## tool ID to directory names mapping, tools preloaded during compilation
use constant CR_TOOL_TO_DIR_MAP	=> {common => 'common', pcc => 'cc', rcc => 'cc'};

## Cache for merged configuration trees.
my %merge_cache;

## returns common prefix for all package messages
sub _msg_prefix {
	my $self = shift;

	return ref $self
		? sprintf('[%s: tool => %s]', __PACKAGE__, $self->{tool})
		: sprintf('[%s]', __PACKAGE__);
}

## Merge all configuration files and check for errors.
## => self
## => cache (unnamed)
## (check $self->{errors} arrayref for errors)
sub _merge {
	my $self = shift;
	my $is_cache = shift;

	my $tool_id = &CR_TOOL_TO_DIR_MAP->{$self->{tool}};
	my $tool_dir = &CR_CONFIG_DIR . '/' . $tool_id;

	## save current dir and change to configuration dir
	unless (-d $tool_dir) {
		push @{$self->{errors}},
			"cannot change to configuration directory ($tool_dir): $!";
		return;
	}

	## enumerate configuration files in alphabetical order
	opendir(DIR, $tool_dir);
	my $filemask = &CR_CONFIG_EXT;
	my @files = sort grep {/\Q$filemask\E$/i} readdir(DIR);
	closedir(DIR);

	## get modification time for each file
	my $merged = {
		tree	=> {},
		_stats	=> {map {$_ => (stat "$tool_dir/$_")[9]} @files},
	};

	## don't actually merge if cache is on or no modifications found since last call
	return
		if $self->{merged} && (
			$is_cache || !grep {
				$self->{merged}->{_stats}->{$_} != $merged->{_stats}->{$_}
			} keys %{$merged->{_stats}}
		);

	## check main configuration file's presence
	my $file_main = shift @files;
	unless ($file_main eq &CR_CONFIG_DEF . &CR_CONFIG_EXT) {
		push @{$self->{errors}},
			'cannot find main configuration file (' . &CR_CONFIG_DEF . &CR_CONFIG_EXT . ')';
		return;
	}

	## prepare containers for content and indexes
	$merged->{$_} = {} ## <ref> for each 'alias' / 'uri'
		foreach qw(_nodes_by_node_id _nodes_by_alias _nodes_by_uri
			_node_containers _node_container_parents);

	## set maximal depth of structure dumps
	local $Data::Dumper::Maxdepth = 1;

	## process files one by one
	my $node_id = 0;
	foreach my $file ($file_main, @files) {
		## check file for existence
		unless (-f "$tool_dir/$file") {
			push @{$self->{errors}},
				"cannot open configuration file ($file): $!";
			return;
		}

		## produce XML tree
		my $xml = eval{
			XMLin("$tool_dir/$file", forcearray => \&CR_CONFIG_NESTING, keyattr => []);
		};
		if ($@) {
			$self->{_logger}->error($self->_msg_prefix . " error parsing $file: $@");
			next;
		}

		## traverse XML tree
		my @to_traverse = ({
			## container element
			parent	=> $xml,
			## structure being read
			to_scan	=> $xml->{&CR_CONFIG_NESTING->[1]},
			## structure being written
			to_fill	=> $merged->{tree},
			## current level
			level	=> 1,
			## URI prefix (concatenation of ancestors' URIs ending with dot)
			uri		=> '',
		});
		my %uri_seen = (); ## track known 'uri' to catch duplicates
		while (my $subtree = pop @to_traverse) {
			## process all nodes of tag and level
			for (my $i = 0; $i <= $#{$subtree->{to_scan}}; ++$i) {
				## element being traversed (can be changed) and original reference
				my $elem = my $elem_orig = $subtree->{to_scan}->[$i];

				my $tag = &CR_CONFIG_NESTING->[$subtree->{level}]
					|| &CR_CONFIG_NESTING->[$#{&CR_CONFIG_NESTING}];

				## add or merge elements
				if (defined $elem->{alias}
						&& defined $merged->{_nodes_by_alias}->{$elem->{alias}}) {
					## merge element with existing one
					my $existing = $merged->{_nodes_by_alias}->{$elem->{alias}};
					my $container = $merged->{_node_containers}->{$existing->{node_id}}
						if defined $existing;
					unless ($container && $container == $subtree->{to_fill}->{$tag}) {
						push @{$self->{errors}},
							"$file: cannot locate 'alias'=\"$elem->{alias}\" "
							. "within nesting level for merging elements:\n"
							. Dumper($elem);
						next;
					}
					## merge (update) attributes and proceed with the merged one
					foreach my $attr (keys %$elem) {
						$existing->{$attr} = $elem->{$attr}
							unless ref $elem->{$attr} eq 'ARRAY'; ## skip children for now
					}
					$elem = $existing;
					$subtree->{parent} = $subtree->{to_fill};
				} elsif (exists $elem->{before}) {
					## process 'before' directive
					my $existing = $merged->{_nodes_by_alias}->{$elem->{before}};
					my $container = $merged->{_node_containers}->{$existing->{node_id}}
						if defined $existing;
					unless ($container && $container == $subtree->{to_fill}->{$tag}) {
						push @{$self->{errors}},
							"$file: cannot locate 'alias'=\"$elem->{before}\" "
							. "within nesting level for inserting element before:\n"
							. Dumper($elem);
						next;
					}
					## move element to the new place
					my $pos;
					for ($pos = 0; $pos <= $#$container && $container->[$pos] != $existing; ++$pos) {};
					splice @$container, $pos, 0, $elem;
					delete $elem->{before};
					$subtree->{parent} = $subtree->{to_fill};
				} else {
					## add new element
					$subtree->{to_fill}->{$tag} ||= [];
					push @{$subtree->{to_fill}->{$tag}}, $elem;
				}

				## check attribute for errors
				foreach my $attr (grep {ref $elem_orig->{$_} ne 'ARRAY'} keys %$elem_orig) {
					unless (exists &CR_NODE_DESCRIPTOR->{$attr}) {
						push @{$self->{errors}},
							"$file: unknown attribute '$attr' of element:\n"
							. Dumper($elem_orig);
						next;
					}
					if ($elem_orig->{$attr} ne ''
							&& ref &CR_NODE_DESCRIPTOR->{$attr} eq 'Regexp'
							&& $elem_orig->{$attr} !~ &CR_NODE_DESCRIPTOR->{$attr}) {
						push @{$self->{errors}},
							"$file: wrong format for '$attr' value of element:\n"
							. Dumper($elem);
						next;
					}
				}

				## assign unique key, 'uri' and 'level' to node, save reference to container
				unless (exists $elem->{node_id}) {
					$elem->{node_id} = ++$node_id;
					$elem->{level} = $subtree->{level};
					my $container = $merged->{_node_containers}->{$elem->{node_id}}
						= $subtree->{to_fill}->{$tag};
					$merged->{_node_container_parents}->{$container}
						||= $subtree->{to_fill};
				}
				$elem->{uri} = $subtree->{uri} . $elem->{id}
					if $elem->{id};

				## process node children
				foreach my $attr (grep {ref $elem_orig->{$_} eq 'ARRAY'} keys %$elem_orig) {
					## check nesting scheme for errors
					unless ($attr eq (&CR_CONFIG_NESTING->[$subtree->{level}+1]
							|| &CR_CONFIG_NESTING->[$#{&CR_CONFIG_NESTING}])) {
						push @{$self->{errors}},
							"$file: wrong nesting for element '$attr' within '$tag':\n"
							. Dumper($elem);
						next;
					}
					## push element for traversing
					push @to_traverse, {
						parent	=> $elem,
						to_scan	=> $elem_orig->{$attr},
						to_fill	=> $elem,
						level	=> $subtree->{level}+1,
						uri		=> substr($elem->{uri} . '.', $[, 6),
								## substr(...) is used to limit URI checks to 3 levels
					};
					delete $elem_orig->{$attr};
				}

				## skip checks and inheritance for merged elements
				next unless $elem == $elem_orig;

				## check for 'uri' and 'alias' uniqueness and create indexes
				foreach (qw(node_id uri alias)) {
					next unless defined $elem->{$_};
					if ($merged->{'_nodes_by_' . $_}->{$elem->{$_}}) {
						push @{$self->{errors}},
							"$file: '$_' is not unique for element:\n"
							. Dumper($elem);
						next;
					}
					$merged->{'_nodes_by_' . $_}->{$elem->{$_}} = $elem;
				}

				## inherit values from ancestors
				foreach my $attr (&CR_ATTR_GET_PARENT) {
					next if defined $elem->{$attr};
					$elem->{$attr} = $subtree->{parent}->{$attr}
						if defined $subtree->{parent}->{$attr};
				}
				## inherit values from 'alias'
				foreach my $attr (&CR_ATTR_GET_ALIAS) {
					next if defined $elem->{$attr};
					$elem->{$attr} = $elem->{alias}
						if defined $elem->{alias};
				}
			}
		}
	}

	## autogenerate missing 'id' and 'uri'
	my @to_traverse = ({
		to_scan	=> $merged->{tree}->{&CR_CONFIG_NESTING->[1]},
		level	=> 1,
		uri		=> '',
		id		=> '01',
	});
	while (my $subtree = shift @to_traverse) {
		my $id_curr = '01';
		foreach my $elem (@{$subtree->{to_scan}}) {
			unless ($elem->{id}) {
				## look for spare 'id' in the nesting level and set 'id' and 'uri'
				for ($elem->{id} = $subtree->{id};
					$merged->{_nodes_by_uri}->{$subtree->{uri} . $elem->{id}};
					++$elem->{id}) {}
			}
			$elem->{uri} = $subtree->{uri} . $elem->{id};
			$merged->{_nodes_by_uri}->{$elem->{uri}} = $elem;

			## check required attributes
			foreach my $attr (&CR_ATTR_REQUIRED) {
				push @{$self->{errors}},
						"required attribute '$attr' absent, and "
						. "nowhere to inherit value from for element:\n"
						. Dumper($elem)
					unless defined $elem->{$attr} && $elem->{$attr} ne '';
			}

			## push element for traversing
			push @to_traverse, {
				to_scan	=> $elem->{&CR_CONFIG_NESTING->[$subtree->{level}+1]
					|| &CR_CONFIG_NESTING->[$#{&CR_CONFIG_NESTING}]},
				level	=> $subtree->{level}+1,
				uri		=> substr($elem->{uri} . '.', $[, 6),
				id		=> $elem->{id},
			};
		}
	}

	## store merged results in cache
	$merge_cache{$tool_id} = $self->{merged} = $merged;

	return;
}

## Return parse tree as XML string.
## => self
## <= xml
sub xml {
	my $self = shift;

	return XMLout($self->{merged}->{tree},
		keyattr => [], rootname => &CR_CONFIG_NESTING->[0])
}

## Create class instance.
## => class
## => tool
## <= self
sub new($%) {
	my $class = shift;
	my %h = (
		tool	=> undef,	## tool ID
		cache	=> 1,		## return instances from cache if found
		@_
	);
	die $class->_msg_prefix . " no or wrong tool passed (tool => '$h{tool}')"
		unless $h{tool} && &CR_TOOL_TO_DIR_MAP->{$h{tool}};

	## prepare object structure
	my $self = {
		_logger	=> HSPC::Logger->new,
		errors	=> [],
		tool	=> $h{tool}, ## current instance's tool ID (for directory name)
	};
	bless $self, $class;

	## try to restore parse tree from cache
	$self->{merged} = $merge_cache{&CR_TOOL_TO_DIR_MAP->{$h{tool}}};

	## merge configuration (first call or file(s) modification); die on errors
	unless ($h{cache} && $self->{merged}) {
		$self->_merge($h{cache});
		## die with report on errors
		die $self->_msg_prefix . "\n" . join('', map {"- $_ \n"} @{$self->{errors}})
			if @{$self->{errors}};
	}

	## clear session data
	$self->reload;

	return $self;
}

## Reload configuration (clear session data).
## => self
## <= self
sub reload($) {
	my $self = shift;

	## clear session data
	$self->{_node_id}		= undef;	## 'node_id' for current node
	$self->{_node_desc_arr}	= {};		## 'node_descriptor': modified nodes (hashref or hashrefs)
	$self->{_node_desc_id}	= undef;	## 'node_descriptor': 'node_id' for node being modified
	$self->{_node_edit_who}	= undef;	## excerpt from call stack for 'node(edit => 1, ...)'

	return $self;
}

## Get 'node_descriptor' on request (uri / alias) or current.
## => self
## => node_id / uri / alias (optional)
## => edit (optional) - 'node_descriptor' is modifiable on 'edit' => 1
## => choose (optional) - set as current node
## <= node_descriptor: hashref with keys from CR_NODE_DESCRIPTOR (see above)
##    || undef on no match
sub node($;@) {
	my $self = shift;
	my %h = (
		node_id	=> undef,
		uri		=> undef,
		alias	=> undef,
		edit	=> 0,
		choose	=> 0,
		@_
	);

	## lookup node on request, if any
	my $node_desc;
	if ($h{node_id} || $h{uri} || $h{alias}) {
		## lookup node
		my $node = undef;
		foreach (qw(node_id uri alias)) {
			$node ||= $self->{merged}->{'_nodes_by_' . $_}->{$h{$_}}
				if $h{$_};
		}
		return undef
			unless $node;

		## set current node ref
		$self->{_node_desc_id} = $node->{node_id};
		## create or restore 'node_descriptor' and fill empty fields
		$node_desc = $self->{_node_desc_arr}->{$self->{_node_desc_id}} ||= {};
		foreach my $attr (keys %$node) {
			$node_desc->{$attr} ||= $node->{$attr};
		}

		## choose node as current, if needed
		$self->{_node_id} = $self->{_node_desc_id}
			if $h{choose};
	} else {
		## get current node if any
		return undef
			unless $self->{_node_id};
		$node_desc = $self->{_node_desc_arr}->{$self->{_node_id}};
	}

	## fill 'node_descriptor' with values of node attributes
	foreach my $attr (keys %{&CR_NODE_DESCRIPTOR}) {
		## already filled and fits format, do not rewrite
		next if defined $node_desc->{$attr}
			&& (ref &CR_NODE_DESCRIPTOR->{$attr} eq 'Regexp'
					&& $node_desc->{$attr} =~ &CR_NODE_DESCRIPTOR->{$attr}
				|| ref $node_desc->{$attr} eq &CR_NODE_DESCRIPTOR->{$attr});
		## not filled yet or erroneous, fill with defaults
		$node_desc->{$attr} = ref &CR_NODE_DESCRIPTOR->{$attr}
			? undef
			: {};
	}

	## run node filter
	if ($node_desc->{filter}) {
		(my $node_package = $node_desc->{class} . '.pm') =~ s!::!/!g;
		eval {
			no strict; ## have to do it to call unknown method from unknown package
			require $node_package;
			return &{$node_desc->{class} . '::' . $node_desc->{filter}}($node_desc); ## static call!
		} or $node_desc->{filtered_out} = 1;
		warn sprintf('[%s] error calling %s::%s: %s', __PACKAGE__,
				$node_desc->{class}, $node_desc->{filter}, $@)
			if $@;
	}

	Encode::_utf8_off($node_desc->{$_})
		foreach keys %$node_desc;

	return $h{edit}
		? $node_desc			## original 'node_descriptor'
		: dclone($node_desc);	## copy of 'node_descriptor'
}

## Get node's 'alias'.
## => (same syntax as 'node' - see above)
## <= alias
##    || undef on no match
sub node_alias($;@) {
	my $self = shift;

	my $node = $self->node(@_)
		or return undef;

	return $node->{alias};
}

## Get 'node_descriptor' of node's parent on request.
## => (same syntax as 'node' - see above)
## => node_descriptor (unnamed, optional)
sub node_parent($;@) {
	my $self = shift;
	my %h = (
		node_id	=> undef,
		uri		=> undef,
		alias	=> undef,
		edit	=> 0,
		choose	=> 0,
		@_
	);

	## try to locate node as parameter or on request
	my $node = ref $_[0] eq 'HASH' && $_[0]->{node_id}
			? shift
			: $self->node(map {$_ => $h{$_}} qw/alias node_id uri/)
		or return undef;

	## try to locate node's parent
	my $parent_id = eval {
		$self->{merged}->{_node_container_parents}->{
			$self->{merged}->{_node_containers}->{$node->{node_id}}
		}->{node_id}
	} or return undef;

	return $self->node(node_id => $parent_id, map {$_ => $h{$_}} qw/edit choose/);
}

## Get node's parent's 'alias'.
## => (same syntax as 'node' - see above)
## <= alias
##    || undef on no match
sub node_parent_alias($;@) {
	my $self = shift;

	my $node = $self->node_parent(@_)
		or return undef;

	return $node->{alias};
}

## Build absolute URI for tool.
## => class / self
## => tool (optional)
## <= link
sub link_tool {
	my $self = shift;
	my %h = (
		tool	=> undef,	## tool ID
		@_
	);
	$self = $self->new(tool => $h{tool})
		unless ref $self; ## new instance

	return join('/', &CR_LINK_LOCATION, $self->{tool});
}


## Build absolute URI for node with optional URI parameters.
## => class / self
## => node_id / uri / alias (optional)
## => ignore (optional) - don't die but return undef of no match
## => URI params (unnamed, optional) - hashref || arrayref
## <= link
##    || undef on no match (in case of 'ignore' flag)
sub link($;@) {
	my $self = ref $_[0]
		? shift ## existing instance
		: shift->new(tool => $ENV{system_obj}->tool_id); ## new instance
	my @uri_params = ref $_[$#_] eq 'ARRAY' ? @{(pop)} : ref $_[$#_] eq 'HASH' ? %{(pop)} : ();

	my %node_lookup = @_; ## node_id / uri / alias

	## lookup node by request or return undef
	my $node = $self->node(%node_lookup)
		or $node_lookup{ignore} and return undef
		or die $self->_msg_prefix . ' node doesn\'t exist ('
			. join(', ', map { "$_ => $node_lookup{$_}" } sort keys %node_lookup)
			. '), caller: ' . join(', ', caller);
	push @uri_params, %{$node->{uri_params}}
		if ref $node->{uri_params} eq 'HASH';

	## process URI params
	my %uri_params_seen;
	for (my $i = 0; $i < @uri_params; ++$i) {
		unless ($uri_params_seen{$uri_params[$i]}) {
			## replace key/value pair with their concatenation
			$uri_params_seen{$uri_params[$i]} = 1;
			splice @uri_params, $i, 2, join(',', $uri_params[$i],
				ref $uri_params[$i+1] eq 'ARRAY'
					? @{$uri_params[$i+1]}
					: $uri_params[$i+1]);
		} else {
			## key already seen, skipping...
			splice @uri_params, $i--, 2;
		}
	}

	return join('/', $self->link_tool, $node->{uri}, @uri_params);
}

## Traverse nodes tree executing callback function for each node - editable!
## => self
## => callback - subref
## => depth_first (optional) - enable depth-first traversal, breadth-first by default
## => max_depth (optional) - traversal depth limitation, no limitation by default
sub traverse($$;%) {
	my $self = shift;
	my $callback = shift;
	my %h = (
		depth_first	=> 0,
		max_depth	=> undef,
		@_,
	);

	my @to_traverse = (map {
		elem	=> $_,
		level	=> 1,
	}, @{$self->{merged}->{tree}->{&CR_CONFIG_NESTING->[1]}});
	my $iter_data = undef;
	while (my $traversing = shift @to_traverse) {
		## execute callback function
		$iter_data =
			$callback->($self->node(node_id => $traversing->{elem}->{node_id}, edit => 1),
				$iter_data);
		## skip to next element on 'max_depth' hit
		next if defined $h{max_depth} && $traversing->{level} >= $h{max_depth};
		## push elements for traversing
		my @to_traverse_next = map {
			elem		=> $_,
			level		=> $traversing->{level}+1,
		}, @{$traversing->{elem}->{&CR_CONFIG_NESTING->[$traversing->{level}+1]
			|| &CR_CONFIG_NESTING->[$#{&CR_CONFIG_NESTING}]}};
		$h{depth_first}
			? unshift @to_traverse, @to_traverse_next
			: push @to_traverse, @to_traverse_next;
		
	}
}

## Create initial instances (preload configuration).
## Useful under mod_perl on 'use HSPC::CompRep' in PerlModule or startup script.
BEGIN {
	return if (caller(2))[0] eq 'Apache::Reload'; ## do not preload on recompile
	__PACKAGE__->new(tool => $_)
		foreach keys %{&CR_TOOL_TO_DIR_MAP};
}

1;

=pod

=head1 NAME

F<CompRep_Implement> stands for B<Parallels Business Automation - Standard Component Repository> I<Implementation> document.

=head1 STATUS

This document contains detailed description of B<Parallels Business Automation - Standard Component Repository>
implementation in technical terms. It is supposed to be used as strict
instruction set for component development and usage.

=head1 IMPLEMENTATION

=head2 CONFIGURATION

=head3 STRUCTURE

B<Parallels Business Automation - Standard Component Repository> I<Configuration> should be presented as
a set of XML documents with the structure described below and specified by
C<comprep_implement.xsd> XML Schema. All configuration files should be located
in one directory and named freely, but with C<.xml> extension and in valid
format in connection with XML Schema mentioned above. Formal validation using
a tool provided must be laid upon every configuration unit before committing it
to B<Parallels Business Automation - Standard Component Repository> I<Configuration>.

B<Parallels Business Automation - Standard Component Repository> I<Configuration> offers the following
vision of B<Parallels Business Automation - Standard> logical structure (in descending order):

=over 4

=item director

First level grouping set with power limited to a reasonable number for main menu
to be observable.

=over

=item manager

Second level grouping set usually based on independent lines of work.

=over

=item screen

Atomic element of structure, any webpage: a dashboard, an edit form,
a wizard step, whatever.

Screens can be nested within other screens to form any URIs and navigation paths.

Screens nested within screens do not have entries in main menu, could be
presented only by higher level nodes.

=back

=back

=back

Refer to C<comprep_implement.xsd> for XML Schema of configuration documents.

Nodes have the following attributes (I<italics> are optional):

=over 4

=item I<C<title_id>> (default: C<alias>)

Localization ID of component's external name; C<string(id =E<gt> 'title_id')>
visually appears in site path, main menu and so on. See note for
C<set_node_context($)> interface below.

=item I<C<alias>>

String unique throughout the configuration; unambiguously identifies node
for direct referencing; used when inserting other nodes on-the-fly,
filling some parameters with default value, etc.

=item I<C<id>>

Number unique on its level; forms C<uri> (as well as I<Screen ID>) as a
concatenation of higher level nodes' C<id> and this C<id>. If C<id> is not set
for a node, it is being set automatically to some of spare values (incrementally).

=item I<C<class>>

Name of handler class.

=item I<C<method>>

Name of handler method from C<class>.

=item I<C<filter>>

Name of I<static> filter function from C<class>.

If present, function is run with C<node_descriptor> (see below) as a parameter
on menu tree creation or screen lookup. B<Should return> non-zero result for
node to be visible and accessible, otherwise node is hidden.

Example of filter function:

 sub filter {
     my $node = shift;
     if ($node->{alias} eq 'node_alias1') {
         return tell_us_whether_to_show_this_node();
     } elsif ($node->{alias} eq 'node_alias2') {
         return get_some_option() == 'some_value';
     }
     return 1;
 }

Value of C<filtered_out> property of C<node_descriptor> is set to C<1> in case
of C<filter> property presented and filter method returning zero result on
call to C<node()>. Thus, other modules don't have to handle C<filter> property
but just check C<filtered_out>.

=item I<C<accounts>>

Account types (C<SW_HSP>, C<SW_RESELLER>, C<SW_CUSTOMER>) as comma separated list -
refer to Security Manager documentation for details.

=item I<C<groups>>

Security groups as comma separated list -
refer to Security Manager documentation for details.

=item I<C<icon>> (default: C<alias>)

Component's page icon.

=item I<C<referrers>>

Screens that may redirect to this screen as comma separated list -
to some cros-links support.

=item I<C<before>>

C<IDREF> to some node's C<alias>; points to node of the same logical level
to insert new node B<before> it.

=item I<C<helper>>

Topic ID of corresponding helper's page.

=item I<C<ssl>>

SSL mode for node; if presented, parameter's value (C<on> or C<off>) specifies
whether to force I<enabled> or I<disabled> SSL for this node (corresponding
"force SSL" option must be enabled in provider's configuration).

=back

Each configuration XML document has C<root> as its root element and nodes of any
level as its children; use C<before> attribute to point to exact place for node
insertion.

All optional parameters are inherited from higher level nodes.

=head3 FILES

=over

=item Location

All configuration files must be located in B<one directory>. Currently it is
C</var/opt/hspc-data/Core/CompRep>.

=item Defaults

Main configuration file must be named C<_.xml>.

=item Order

Files are being sorted in B<alphabetical order> before reading.
Use it for organizing sequences of dependent configuration blocks.

=item Inheritance

To add a record to a block, reproduce the whole nesting scheme of the block,
using the only C<alias> parameter for matching of parent nodes, and place
the record inside the scheme. If a record with such C<alias> already exists,
it gets B<replaced>. Use it for overriding blocks of configuration.

=back

=head3 LINKS

Node URIs are concatenation of C<id> on each nesting level.
B<Parallels Business Automation - Standard Component Repository> provides an interface for building absolute
URLs for nodes, so B<don't hard code links>, use C<link()> interface instead.

=head2 INTERFACES

B<Parallels Business Automation - Standard Component Repository> is implemented in C<HSPC::CompRep> package
and provides following interfaces (parameters in B<bold> are required):

=over 4

=item C<xml()>

B<Returns> configuration parse tree as XML string.

Always remember this handy one-liner:

 perl -MHSPC::CompRep -e "print HSPC::CompRep->new(tool => 'TOOL')->xml"

Use it to check your configuration by replacing C<TOOL> with the needed
(C<common>, C<pcc>, C<rcc>, etc).

=item C<new()>

B<Parameters:> B<C<tool>>, C<cache>

Check configuration of C<tool> for validity, accuracy and consistency, merge to
single XML and build parse tree, unless object instance is not cached, otherwise
get cached object; C<die> on errors while merging.

Non-null (default) value of C<cache> enables C<new()> to return results
preloaded on server startup instead of merging configuration on each call.
Otherwise configuration is completely reloaded on each call to C<new()>,
if any or some of configuration files have changed since previous call.

B<Returns> blessed reference (C<$self>) as the result.

=item C<reload()>

Nullify current node pointer and clear modifications made to configuration
since call to C<new()>.

=item C<node()>

B<Parameters:> C<alias>, C<node_id>, C<uri>, C<edit>, C<choose>

If B<one or several> request parameters (C<alias>, C<node_id>, C<uri>) are passed,
lookup correspondent node (in B<OR> mode). If B<no> parameters are passed,
lookup B<current> node.

Find node by C<alias>:

 $cr->node(alias => 'node_alias');

Find node by C<uri>:

 $cr->node(uri => '01.01.01');

 $cr->node(uri => '01.01');

Find node by C<alias> B<OR> C<uri> (returns last matching):

 $cr->node(alias => 'node_alias', uri => '01.01.01');

Find current node:

 $cr->node;

B<Returns> C<node_descriptor>, which is a reference to hash with all node
attributes as keys plus C<title_hash> and C<uri_params>, or C<undef> on no match.

=over

=item *

C<title_hash> is hashref with parameter substitutions for C<title_id>:

 $cr->node(edit => 1)->{title_hash}->{param} = 'VALUE';
 ## i18n "Node #%param%" turns into "Node VALUE" in titles, menus, etc.

=item *

C<uri_params> is hashref with URI parameters added to node's link:

 print $cr->link;                                ## /01
 $cr->node(edit => 1)->{uri_params}->{param} = 'VALUE';
 print $cr->link;                                ## /01/param,VALUE
 print $cr->link([param2 => 'VALUE2']);          ## /01/param,VALUE/param2,VALUE2

=back

Nodes modification is done using techniques described below.

=over

=item *

If non-zero C<edit> parameter is passed, C<node_descriptor> is B<modifiable>:
all modifications applied to C<node_descriptor> keys affect the internal
representation of configuration and, therefore, the final output.
It is useful for dynamic generation of page parameters, for example.
Otherwise C<node_descriptor> is a clone of node structure, which could
be modified without affecting configuration (I<default>).

=item *

If non-zero C<choose> parameter is passed, C<node()> not only returns
C<node_descriptor>, but sets the chosen node as I<current>, i.e. the one
which is returned on no parameters call to C<node()> and used in web page
generation. Otherwise node is not set as current (I<default>).

=back

Note that C<node()> doesn't die on lookup failure, it just returns C<undef>,
so don't hesitate to do as many lookups as you need - to check a plug-in's
configuration, for example.

Use chains of C<node()> calls with non-zero C<edit> parameter to define all
undefined parts of path, if any (parameters substitution for title localization
- C<title_hash> - or whatever).

Use call to C<node()> with non-zero C<choose> parameter to explicitly set
current node before producing the page output.

=item C<node_alias()>

B<Parameters:> C<alias>, C<node_id>, C<uri>

Try to locate node on request and return its C<alias>.

B<Returns> node's C<alias>.

=item C<node_parent()>

B<Parameters:> C<alias>, C<node_id>, C<uri>, C<edit>, C<choose>, I<C<node_descriptor>>

If B<one or several> request parameters (C<alias>, C<node_id>, C<uri>) are passed,
lookup correspondent node's B<parent> (in B<OR> mode). If B<no> parameters are passed,
lookup B<current> node's B<parent>.

If an unnamed I<C<node_descriptor>> parameter is passed, do not lookup node
on request but use this parameter's value as node to lookup parent for.

Find node's parent by node's C<alias>:

 $cr->node_parent(alias => 'node_alias');

Find node's parent by node's C<node_descriptor>:

 my $node = $cr->node(alias => 'node_alias');
 $cr->node_parent($node);

B<Returns> C<node_descriptor> or C<undef> on no match.

=item C<node_parent_alias()>

B<Parameters:> C<alias>, C<node_id>, C<uri>

Try to locate node's parent's on request and return its C<alias>.

B<Returns> node's parent's C<alias>.

=item C<link_tool()>

B<Parameters:> I<C<tool>>

Produce absolute link corresponding to current or specified tool.

B<Returns> tool's absolute link.

=item C<link()>

B<Parameters:> C<node_id>, C<alias>, C<uri>, C<ignore>, I<C<uri_params>>

Lookup node by passing C<node_id> or C<alias> / C<uri> (both if present) to
C<node()> and produce absolute link corresponding to the node. I<C<uri_params>>
is an unnamed parameter, expected just after C<alias> and/or C<uri>, if any,
and is a hashref B<or> arrayref of URI parameters. B<Note:> I<C<uri_params>>
B<overrides> URI params already presented in C<uri_params> field of
C<node_descriptor>.

Find link by C<alias>, include C<state> URI parameter:

 $cr->link(alias => 'some_node', {state => 'ok'});

Find link by C<alias>, include C<state> and C<id> URI parameters B<considering> order:

 $cr->link(alias => 'some_node', [state => 'ok', id => 13]);

Find link by C<alias>, include C<id> URI parameter B<overriding> C<uri_params>:

 $cr->node(alias => 'some_node', edit => 1)->{uri_params}->{id} = 1;
 $cr->link(alias => 'some_node', {id => 1});

Same as above, but using weird technique:

 $cr->node(alias => 'some_node', choose => 1);
 $cr->link({state => 'ok'});

Find link for current node:

 $cr->link;

B<Returns> node's absolute link, dies on no match and C<ignore> being Perl's B<false>
or C<undef> on no match and C<ignore> being Perl's B<true>.

=item C<traverse()>

B<Parameters:> B<I<C<callback>>>, C<depth_first>, C<max_depth>

Traverse nodes tree executing callback function passed as a parameter for each
node - editable. Unnamed parameter I<C<callback>> is a reference to subroutine,
which is being called for each node of the nodes tree with two unnamed
parameters: I<C<node>> (C<node_descriptor> structure) and I<C<iter_data>>
(the result of previous call of callback function). Optional parameters
C<depth_first> and C<max_depth> specify correspondently whether nodes tree
should be traversed in depth-first fashion (breadth-first fashion is used
otherwise) and what is maximum depth to dig.

Scan all nodes depth-first, enumerating and printing C<uri>:

 $cr->traverse(sub {
    my $node = shift;
    my $iter = shift || {index => 1};
    print "$iter->{index} $node->{uri}\n";
    return {index => $iter->{index} + 1};
 }, depth_first => 1, max_depth => 3);

=back

Public attributes:

=over 4

=item C<tool>

Tool ID for the class instance (e.g. C<common>, C<pcc>, C<rcc>, etc.).
Appended to global configuration path in the filesystem gives a directory
with configuration specific to the tool.

=back

=head3 mod_perl

C<HSPC::CompRep> is preloaded in C<BEGIN> block for all supported tools
(see C<HSPC::CompRep::CR_CONFIG_TOOLS> constant) on first C<use> directive:

 use HSPC::CompRep;

Including this directive in C<PerlModule> or I<startup script> of C<mod_perl>
configuration makes heavy repository data get loaded in parent process and
shared among children. C<HSPC::Startup> contains such C<use> directive.

Calls to C<new()> B<do not reload> configuration, but return the
B<preloaded object>. Feel free to call C<new()> in each module that needs to
access repository to get the object without hitting system performance.

=head2 EXAMPLE

The following example demonstrates configuration comprised of three XML documents.

=over

=item *

C<_.xml>

Define:

=over

=item C<director> C<account_director>

=item C<manager> C<customer_manager> in C<account_director>

=item C<screen> C<customers> in C<customer_manager>

=back

XML:

 <root>
    <director alias="account_director" title_id="account_director"
              class="HSPC::CC::Dashboards" method="dashboard_acd"
              icon="account_properties" helper="managing_accounts.htm">
       <manager alias="customer_manager" title="Customer Manager" id="05"
                class="HSPC::CC::Dashboards" method="dashboard_bd_am"
                icon="account_manager">
          <screen alias="customers" title="Customers" id="01"
                  class="HSPC::AM::CCAccounts" method="teaser_accounts"
                  icon="users">
             <screen alias="customers_new" title="New Account" id="02" />
          </screen>
       </manager>
    </director>
 </root>

=item *

C<reseller_customers.xml>

Define:

=over

=item C<manager> C<reseller_manager> in C<account_director>

=item C<screen> C<reseller_customers_new> in C<reseller_manager>

=back

XML:

 <root>
    <director alias="account_director">
       <manager alias="reseller_manager" title="Reseller Manager" id="06"
                class="HSPC::CC::Dashboards" method="dashboard_bd_cm"
                icon="reseller_manager">
          <screen alias="reseller_customers_new" title="New Account" id="02"
                  class="HSPC::AM::CCAccounts" method="teaser_accounts"
                  icon="users"/>
       </manager>
    </director>
 </root>

=item *

C<reseller_customers_new.xml>

Define:

=over

=item C<screen> C<reseller_customers> in C<reseller_manager> before C<reseller_customers_new>

=back

XML:

 <root>
    <director alias="account_director">
       <manager alias="reseller_manager">
          <screen alias="reseller_customers" title="Customers" id="01"
                  before="reseller_customers_new"
                  class="HSPC::AM::CCAccounts" method="teaser_accounts"
                  icon="users"/>
       </manager>
    </director>
 </root>

=item *

XML returned by C<merge()> - merged result:

 <root>
    <director alias="account_director" title="Account Director" id="05"
              class="HSPC::CC::Dashboards" method="dashboard_acd"
              icon="account_properties" helper="managing_accounts.htm">
       <manager alias="customer_manager" title="Customer Manager" id="05"
                class="HSPC::CC::Dashboards" method="dashboard_bd_am"
                icon="account_manager">
          <screen alias="customers" title="Customers" id="01"
                  class="HSPC::AM::CCAccounts" method="teaser_accounts"
                  icon="users">
             <screen alias="customers_new" title="New Account" id="02" />
          </screen>
       </manager>
       <manager alias="reseller_manager" title="Reseller Manager" id="06"
                class="HSPC::CC::Dashboards" method="dashboard_bd_cm"
                icon="reseller_manager">
          <screen alias="reseller_customers_new" title="New Account" id="02"
                  class="HSPC::AM::CCAccounts" method="teaser_accounts"
                  icon="users" />
          <screen alias="reseller_customers" title="Customers" id="01"
                  before="reseller_customers_new"
                  class="HSPC::AM::CCAccounts" method="teaser_accounts"
                  icon="users" />
       </manager>
    </director>
 </root>

=back

=head1 AUTHOR

Nikita Dedik E<lt>ndedik@sw.ruE<gt>

=head1 COPYRIGHT

Copyright 2005 SWsoft

=cut
