# Copyright (C) 2000, 2001  SWsoft
# All rights reserved.
#
package HSPC::MT::Abstract;

use strict;

use SWsoft::protect;
use HSPC::Localization qw(lang_field_select_expr current_lang);
use HSPC::WebDB qw(:DEFAULT trans trans_counter);
use HSPC::System;

use HSPC::Logger qw(sw_log_debug sw_die);
use HSPC::DataRequest;

##use HSPC::MT::Core::ObjType;
use HSPC::MT::LangValue;

use vars qw($AUTOLOAD $DEBUG);

############################################################################
# accessor methods
sub AUTOLOAD {
    my $self = shift;
    my $class = ref($self);
    return unless $class;
    my $attr = $AUTOLOAD;
    $attr =~ s!.*::!!;
				return if ( $attr =~ /^[A-Z]*$/ ); ## skip DESTROY and all-cap methods
    my $props = $self->_get_class_info->{props};
    my $prop = $props->{$attr};
	if($prop) {
		if( $prop->{type} && $prop->{type} eq 'lang' && !defined $self->{$attr} ) {
			$self->{$attr} = new HSPC::MT::LangValue();
		}
		if (@_) {
			sw_die("ERROR: updating read-only attribute is prohibited: $attr for $class object")
				if ($prop->{not_save} || $prop->{read_only});
			my $value = shift;
			$self->is_mt_obj_changed(1) if ( $value ne $self->{$attr} );
			# check translated
			if($prop->{type} eq 'version') {
				sw_die("ERROR: Don't set object version! It's internal data of Abstract!")
			}
			if($prop->{type} eq 'translate') {
				sw_die("ERROR: value $value for translated attribute $attr is incorrect")
					unless scalar grep $_ eq $value, values %{$prop->{translate}};
			}
			if ( $prop->{type} && $prop->{type} eq 'lang') {
				eval {
					$self->{$attr}->set_hash($value);
				};
				sw_die($@) if ( $@ );
			} else {
				$self->{$attr} = $value;
			}
		}
		# load list attribute if not loaded yet
		if($prop->{type} && $prop->{type} eq 'list' && !defined $self->{$attr}) {
			my @defined_keys = grep { $props->{$_}->{key} && $self->{$_} } keys %$props;
			_operate_list($self,$attr,'load') if scalar @defined_keys;
			$self->{$attr} = [] unless defined $self->{$attr};

			## convert to class if necessary
			if ($prop->{class_ref}) {
				my @new_collection;
				foreach my $item ( @{ $self->{$attr} } ) {
					if ( ref($item) eq 'HASH' ) {
						push @new_collection, $prop->{class_ref}->new(%$item);
					} else {
						push @new_collection, $item;
					}
				}
				$self->{$attr} = \@new_collection;
			}
		}
		return $self->{$attr};
	} else {
		sw_die("Can't find method $class -> $attr");
	}
}

############################################################################
# helper function to generate SQL CASE expression
sub _translate_case {
	my ($name,$trans) = @_;
	my $str = ' case '.$name;
	foreach (keys %$trans) {
		$str .= ' when '.HSPC::WebDB::quote($_).' then '.HSPC::WebDB::quote($trans->{$_});
	}
	return $str.' end ';
}


############################################################################
# > class_info => $class_info # reference on hash with class data
# < class item reference
sub new {
	my $class = shift;
	my %h     = @_;
	my $explicit_class_info = $h{explicit_class_info};

	my $self = {};
	bless ($self,$class);

	my $class_info = _get_class_info($class, $explicit_class_info);

	$self->{_webdb_} = new HSPC::WebDB(dbname => $class_info->{database});
	$self->{_is_mt_obj_changed} = 0;

	return $self;
}

############################################################################
# > class - name of the class, for which CLASS_INFO will be collected
# < class_info - reference to CLASS_INFO structure
my %ci_cache;
sub _get_class_info {
	my $this            = shift;
	my $cstm_class_info = shift;

	my $class = ref($this) || $this;
	my $class_info = $ci_cache{$class};

	unless ($class_info) {
		$class_info = _collect_class_info($class, $cstm_class_info);
		sw_die("CLASS_INFO was not found!") unless $class_info;

		# check CLASS_INFO validity
		my $err = _check_class_info($class_info);
		sw_die("$class CLASS_INFO is invalid: $err") if $err;

		$ci_cache{$class} = $class_info;
	}
	return $ci_cache{$class};
};

############################################################################
# > class - name of the class, for which CLASS_INFO will be collected
# < class_info - reference to CLASS_INFO structure
sub _collect_class_info {
	my $class = shift;
	my $class_info = shift || { props=>{}, tables=>{}, level=>{} };
	my $level = shift || 1;

	## cut off recursion
	if ($ci_cache{$class}) {
		_merge_class_info($class_info, $ci_cache{$class}, $class, $level);
		return $class_info;
	}

	my $cur_ci = __get_constant($class, 'CLASS_INFO');;
	$cur_ci->{database} = 'MAIN' if (!$cur_ci->{database});

	# if got CLASS_INFO
	if ($cur_ci && ref($cur_ci) eq 'HASH') {
		# merge it with CLASS_INFO
		_merge_class_info($class_info,$cur_ci,$class,$level);
	}

	# go through inheritance tree recursively
	my $name  = $class . '::ISA';
	my @parents;
	{
		no strict 'refs';
		@parents = @{$name};
	}
	++$level;
	foreach (@parents) {
		next if $_ eq __PACKAGE__;
		_collect_class_info($_, $class_info, $level);
	}
	return $class_info;
}

sub __get_constant {
	my $class = shift;
	my $name = shift;
	# try to get CLASS_INFO from package
	my $name = $class . '::' . $name;
	no strict 'refs';
	return *{$name}{CODE} ? &$name : undef;
}

############################################################################
# primarily used to verify CLASS_INFO structure after merging
#
# > main_class_info - main CLASS_INFO structure
# > sub_class_info - CLASS_INFO structure to be added to main one
sub _merge_class_info {
	my ($main_ci,$sub_ci,$sub_class,$level) = @_;
	
	# check database 
	if (defined $main_ci->{database}) {
		sw_die("Cannot merge classes, using different databases")
			if ($main_ci->{database} ne $sub_ci->{database});
	} else {
		$main_ci->{database} = $sub_ci->{database};
	}

	{ ## lock_on_load attribute merging.
		if (defined $sub_ci->{lock_on_load}
			&& !exists($main_ci->{lock_on_load})
		) {
			$main_ci->{lock_on_load} = $sub_ci->{lock_on_load}
		}
	}
	
	{ ## no_trans attribute merging.
		if (defined $sub_ci->{no_trans}
			&& !exists($main_ci->{no_trans})
		) {
			$main_ci->{no_trans} = $sub_ci->{no_trans}
		}
	}
	

	{ ## no_database attribute merging.
		if (defined $sub_ci->{no_database}
			&& !exists($main_ci->{no_database})
		) {
			$main_ci->{no_database} = $sub_ci->{no_database}
		}
	}

	{ ## group_by attribute merging.
		if (defined $sub_ci->{group_by}
			&& !exists($main_ci->{group_by})
		) {
			$main_ci->{group_by} = $sub_ci->{group_by};
		}
	}

	# copy additional properties
	my ($main_props,$sub_props) = ($main_ci->{props},$sub_ci->{props});

	# go through properties list
	foreach (keys %$sub_props) {
		# skip attribute if it's already exists
		next if exists $main_props->{$_};
		
		# local property hash
		sw_die("something wrong with class info sub attribute: '$_'")
			unless $sub_props->{$_};

		my %prop = %{$sub_props->{$_}};

		# if it's list attribute, save CLASS_INFO structure
		if (exists $prop{type} && $prop{type} eq 'list' && !$prop{_class_info_}) {
			if ($prop{class_info}) {
				my $list_ci = __get_constant($sub_class, $prop{class_info});
				die qq{ERROR: no $prop{class_info} structure defined for $sub_class} unless $list_ci;
				$prop{_class_info_} = $list_ci;
			} elsif ($prop{class_ref}) {
				my $list_ci = _get_class_info($prop{class_ref});
				die qq{ERROR: can't load class info for subclass $sub_class} unless $list_ci;
				$prop{_class_info_} = $list_ci;
			} else {
				sw_die("Neither class_info nor class_ref specified");
			}
		}

		# add property
		$main_props->{$_} = \%prop;
	}

	# copy additional tables
	my ($main_tables,$sub_tables) = ($main_ci->{tables},$sub_ci->{tables});
	my @new_tables = grep !exists($main_tables->{$_}), keys %$sub_tables;
	if (exists $sub_ci->{level}) {
		foreach (@new_tables) {
			$main_ci->{level}{$_} = $sub_ci->{level}{$_} + $level - 1;
		}
	} else {
		foreach (@new_tables) {
			$main_ci->{level}{$_} = $level;
		}
	}
	@{$main_tables}{@new_tables} = @{$sub_tables}{@new_tables};
 
}

############################################################################
# > class_info - CLASS_INFO structure to check
# < error string if CLASS_INFO is not valid
sub _check_class_info {
	my $class_info = shift;

	my ($props,$tables) = ($class_info->{props},$class_info->{tables});

	unless ($class_info->{no_database}) {
		# check CLASS_INFO validity
		my @mt = grep !exists $tables->{$_}->{rel_table}, keys %$tables;
		if(scalar @mt > 1) {
			return "CLASS_INFO has more than one base table";
		} elsif(scalar @mt == 0) {
			return "CLASS_INFO don't has a base table";
		}

		unless(scalar keys %$props > 0) {
			return "CLASS_INFO has no attributes";
		}
	}

	return;
}

############################################################################
# > id_col_name => $id # item id value
# > debug (1/0, 0 by default)
# => read_only: load object for read_only purposes.
sub load {
	my $self = shift;
	my %h = (
		load_lists	=> undef,
		read_only	=> undef,
		@_
	);

	my $ci = $self->_get_class_info;
	my $props = $ci->{props};

	sw_die("try to load object, which not designed to use database")
		if $ci->{no_database};
	
	

	# turn debug mode on/off
	local $DEBUG = $h{debug} || 0;

    # load object data
   	foreach (keys %$self) {
   		$h{$_} = $self->{$_} unless exists $h{$_};
   	}

	## define - do we need to lock object?
	my $need_lock = 0;
	if ($ci->{lock_on_load}		## we need locking for this class.
		&& !$h{read_only} 		## we load object not in r/o mode.
		&& trans_counter() > 0	## we inside transaction. Lock outside has no meaninig.
	) {
		$need_lock = 1;
	}

	my $data = _load($self,$ci,\%h,0,$need_lock);

	# clear any values stored in object
	foreach (keys %$props) {
		undef $self->{$_};
	}

	# store data in object #
	foreach (grep {$props->{$_}->{type} ne 'lang'} keys %$data) {
		$self->{$_} = $data->{$_};
	}
	foreach (grep {$props->{$_}->{type} eq 'lang'} keys %$data) {
		$self->{$_} = new HSPC::MT::LangValue();
		$self->{$_}->from_db_format($data->{$_});
	}

	# load list properties
	if($h{load_lists}) {
		foreach my $prop_name ( grep exists $props->{$_}->{type} &&
			$props->{$_}->{type} eq 'list', keys %$props ) {
			# load data for list attribute
			_operate_list($self,$prop_name,'load');
		}
	}

	## set flag, that this object was load for read_only purposes.
	$self->{__object_was_load_as_read_only} = $h{read_only};
	## set flag, if object was locked
	$self->{__object_was_locked} = $need_lock;
	
	

	return;
}

############################################################################
## this checks - is object loaded with lock?
sub is_object_locked {
	my $self = shift;
	return $self->{__object_was_locked};
}

############################### PRIVATE ####################################
# > ref to object
# > attribute name
sub _operate_list {
	my $obj = shift || die;
	my $prop_name = shift || die;
	my $operation = shift || die;
	
	my $prop = $obj->_get_class_info->{props}->{$prop_name};
	die qq{ERROR: property $prop_name is not a list attribute}
		unless $prop->{type} eq 'list';

	my $subci = $prop->{_class_info_};
	die qq{ERROR: no class_info found for $prop_name attribute} unless $subci;
	die qq{ERROR: no join specified for $prop_name attribute} unless $prop->{join};

	if($operation eq 'load') {
		# prepare attribute values for child objects
		my $vals = {};
		@{$vals}{keys %{$prop->{join}}} = map { $obj->{$_} } values %{$prop->{join}};

		# load list
		my $data = _load($obj,$subci,$vals,1,0);
		$obj->{$prop_name} = $data || [];
		foreach my $r ( @{$data} ) {
			foreach my $q (grep {$subci->{props}->{$_}->{type} eq 'lang'} keys %{$r}) {
				my $val = new HSPC::MT::LangValue();
				$val->from_db_format($r->{$q});
				$r->{$q} = $val;
			}
		}

	} elsif($operation eq 'save') {
		# save list
		if (!$obj->{$prop_name}
			|| ref($obj->{$prop_name}) ne 'ARRAY'
		) {
			sw_die("$prop_name should be array reference");
		}
		
		# delete all records first
		my %parent;
		foreach (keys %{$prop->{join}}) {
			$parent{$_} = $obj->{$prop->{join}->{$_}};
		}
		_delete($obj,$subci,\%parent,1);

		# then save each record (one-by-one)
		sw_die("$prop_name should be array reference") unless (
			$obj->{$prop_name} && ref($obj->{$prop_name}) eq 'ARRAY'
		);
		foreach my $row (@{$obj->{$prop_name}}) {
			# make sure that it has parent attributes
			@{$row}{keys %parent} = values %parent;
			# save child record
			_save($obj,$subci,$row,1);
		}

	} elsif($operation eq 'delete') {
		# delete all records
		my %parent;
		foreach (keys %{$prop->{join}}) {
			$parent{$_} = $obj->{$prop->{join}->{$_}};
		}
		_delete($obj,$subci,\%parent,1);
	}
}


################################ PRIVATE ###################################
# > CLASS_INFO
# > ref to hash with attribute values
# > whether all keys should be specified (1 by default)
sub _load {
	my $obj = shift;
    my $ci = shift;
    my $vals = shift;
    my $list = shift || 0;
    my $need_lock = shift || 0;

    # init vars #
    my @cols=();
    my @tables=();
	my @joins=();
	my @where=();
	my $props = $ci->{props};
	my $tables = $ci->{tables};

	# process properties #
	foreach my $prop_name ( keys %$props ){
		my $prop = $props->{$prop_name};

		# skip this attribute if it's write only (f.e. password)
		next if $prop->{write_only};

    	# construct column SQL code #
    	my $column;
		if(exists $prop->{type} && $prop->{type} eq 'translate') {
			# if column value should be translated into attribute value
			$column = _translate_case(
				$prop->{table}.'.'.$prop->{col},$prop->{translate}
			);
		} else {
			if(exists $prop->{sql}) {
				# if column expressed via custom SQL code
				$column = $prop->{sql};
			} elsif(exists $prop->{col}) {
				# ordinary column definition (table + column)
				$column = $prop->{table}.'.'.$prop->{col};
			} else {
				# non-typical column
				next;
			}
		}
		# add column into columns list
		push(@cols, $column.' AS '.$prop_name);

		if( (exists $vals->{$prop_name}) && ($list || $prop->{key})) {
			if(ref($vals->{$prop_name}) eq 'ARRAY') {
				push(@where, $column.' IN ('.
					join(',',map { HSPC::WebDB::quote($_) } @{$vals->{$prop_name}}).')');
			} else {
				push(@where,$column.( defined($vals->{$prop_name}) ? 
					'='.HSPC::WebDB::quote($vals->{$prop_name}) : ' is NULL'));
			}
		} elsif($prop->{key} && !$list) {
			sw_die("ERROR : key attribute $prop_name is empty ($vals->{$prop_name}), cant load data $obj");
		}
	}

	# handle tables #	
	
	# order tables so main table will be first,
	# then will appear tables joined to main table
	# and then all other tables
	# prepare tables for SQL
	my ($tmain_alias) = grep !exists $ci->{tables}->{$_}->{rel_table}, keys %$tables;
	my @tabs = ($tmain_alias, 
		(grep $tables->{$_}->{rel_table} && 
			$tables->{$_}->{rel_table} eq $tmain_alias, sort { $tables->{$a}->{order} <=> $tables->{$b}->{order} } keys %$tables),
		(grep $tables->{$_}->{rel_table} && 
			$tables->{$_}->{rel_table} ne $tmain_alias, sort { $tables->{$a}->{order} <=> $tables->{$b}->{order} } keys %$tables));
	# prepare tables for SQL
	foreach my $t_name ( @tabs ){
		my $t = $ci->{tables}->{$t_name};
		my $name = $t->{name}.' '.$t_name;

		if( $t->{join_where} ){
			push(@tables, $name);
			push(@where, qq{( $t->{join_where} )} );
		}elsif( $t->{join_inner} ){
			die unless $t->{rel_table};
			push(@joins, qq{ INNER JOIN $name ON $t->{join_inner} });
		}elsif( $t->{join_left} ){
			die unless $t->{rel_table};
			push(@joins, qq{ LEFT JOIN $name ON $t->{join_left} });
		}elsif( $t->{join_right} ){
			die unless $t->{rel_table};
			push(@joins, qq{ RIGHT JOIN $name ON $t->{join_right} });
		}else{
			die "No relation between table $t_name and $t->{rel_table}" if $t->{rel_table};
			push(@tables, $name);
		}
	}

	# prepare SQL-command #
	my $sql='SELECT '.join(',',@cols).
			' FROM '.join(',',@tables).join('',@joins);
	$sql .= ' WHERE '.join(' AND ',@where) if scalar @where;
	$sql .= ' GROUP BY '.$ci->{group_by} if ( $ci->{group_by} );

	if ($need_lock) {
		$sql .= ' FOR UPDATE';
	}

	# run SQL-command #
	return $list ? $obj->{_webdb_}->select_hashrows($sql) : $obj->{_webdb_}->select_hash($sql);
}

############################################################################
# < undef || errors_string # string with errors messages
sub save {
	my $self = shift;
	my %h = (@_);

	## check - may be we try to save object, which we load as read_only?
	sw_die("try to save object, which was load as 'read_only'")
		if ($self->{__object_was_load_as_read_only});

	# check object for consistency
	my $errors = $self->check(%h);
	return $errors if $errors;

	# turn debug mode on/off
	local $DEBUG = $h{debug} || 0;

	my $ci = $self->_get_class_info;
	my $props = $ci->{props};

	sw_die("try to save object, which not designed to use database")
		if $ci->{no_database};

	# start transaction
	my $trans = !$ci->{no_trans} ? trans(comment=>"Abstract->save", webdb=>$self->{_webdb_}) : undef;

	# save object
	_save($self,$ci,$self);

	# handle list properties
	foreach my $prop_name (grep exists $props->{$_}->{type} &&
		$props->{$_}->{type} eq 'list' &&
		!$props->{$_}->{not_save} && !$props->{$_}->{read_only}, sort keys %$props) {
		# save data for list attribute
		_operate_list($self,$prop_name,'save') if defined $self->{$prop_name};
	}
    $trans->commit() if !$ci->{no_trans};
	
	return;
}

################################ PRIVATE ###################################
# > CLASS_INFO
# > ref to hash with attribute values
sub _save {
	my $obj = shift;
    my $ci = shift;
    my $vals = shift;
    my $list = shift || 0;

	my $col;
	my $prop;	

	my %tables = %{$ci->{tables}};
	my %props = %{$ci->{props}};
	
	my %auto_vals=(); # AUTO property values for direct-joined tables

	# get name of the the main table
	my ($tmain_alias) = grep !exists $tables{$_}->{rel_table}, keys %tables;
	unless($tmain_alias){		
		die q[Main table was not found for class];
	}

	# get list of all direct-joined tables
	# (skip table if it's read-only)
	my @joined_aliases = grep exists $tables{$_}->{rel_table} &&
		$tables{$_}->{rel_table} eq $tmain_alias && 
		!$tables{$_}->{not_save} && !$tables{$_}->{read_only} &&
		$tables{$_}->{join_where}, sort keys %tables;

	# add new properties for hard-joined tables
	_extend_properties($ci,\%props,$vals) if scalar @joined_aliases;

	my @sorted_joined_aliases;
	if ( scalar @joined_aliases ) {
		my %levels;
		if ( exists $ci->{level} && defined $ci->{level} ) {
			%levels = %{$ci->{level}};
		} else {
			sw_die("Script must not be here. This code is for foreign keys implementation. Ask billing team.");
		}
		my @sorted_tables = sort { $levels{$b} <=> $levels{$a} } keys %levels;

		foreach my $table (@sorted_tables) {
			if ( scalar grep { $table eq $_ } @joined_aliases ) {
				push(@sorted_joined_aliases, $table);
			}
		}
	}

	# update each table
	foreach my $t_alias ($tmain_alias,@sorted_joined_aliases) {
		# table properties
		my $t = $tables{$t_alias};
		# table real name
		my $t_name = $t->{name};

		# attributes to be updated:
		# + attribute should be contained in main table
		# + attribute should not be 'not_save'/'read_only' (when updating)
		# + attribute should be defined as an ordinary column
		my @prop_names = grep exists $props{$_}->{table} && 
			$props{$_}->{table} eq $t_alias &&
			(!$props{$_}->{not_save} && !$props{$_}->{read_only}) &&
			$props{$_}->{col},
			keys %props;
			
		# If no attributes to be updated
		return unless scalar @prop_names;

		# die if there are no key attributes
		unless (scalar (grep $props{$_}->{key}, @prop_names) > 0) {
			die qq{ERROR: cannot save record with no key attributes, table = $t_alias};
		}

		# die if some key is not defined
		my @empty_props = (grep $props{$_}->{key} && $props{$_}->{key} ne 'AUTO' && 
			!defined $vals->{$_} && !$auto_vals{$_}, @prop_names);
		if (scalar (@empty_props) > 0) {
			sw_die(
				'ERROR : cannot save props '.join(',',@empty_props)."\n"
				.'ERROR : cannot save record when non-AUTO attribute is undefined'
			);
		}

		# check if we gonna create new record or update existent
		my $new_record = $t->{replace} || $list || 
			scalar (grep exists $props{$_}->{key} &&
			$props{$_}->{key} eq 'AUTO' &&
			!$vals->{$_}, @prop_names) > 0 ? 1 : 0;

		my %change_props;

		if($new_record){
			# CREATE NEW RECORDS #
			my @cols=();
			my @values=();
			my $prop_name;
			my $prop_auto;

			# process key_props #
			foreach $prop_name (@prop_names) {
	    		$prop = $props{$prop_name};

				# if we inserting in main table - remember AUTO property
				if($t_alias eq $tmain_alias && exists $prop->{key} && 
					$prop->{key} eq 'AUTO' && !$vals->{$prop_name}) {
					$prop_auto = $prop_name;
					next;
				}


				# add column name into columns list
				push(@cols, $prop->{col});

				# by default we use quoted attribute value
				my $value = "''";

				if ($prop->{type} eq 'version') {
					## set version of new object to 1.
					$value = 1;
					$change_props{$prop_name} = $value;
				} elsif (exists $prop->{type} && $prop->{type} eq 'lang' ) {
					$value = HSPC::WebDB::quote($vals->{$prop_name}->to_db_format()) if (
						ref($vals->{$prop_name})
					);
				} else {
					$value = HSPC::WebDB::quote(
						exists $auto_vals{$prop_name} ? 
						$auto_vals{$prop_name} : $vals->{$prop_name}
					);
				}
				# translate value if needed
				$value = _translate_case($value,{ reverse %{$prop->{translate}} })
					if exists $prop->{type} && $prop->{type} eq 'translate';

				# add it in values list
				push(@values, $value);
			}

			unless (scalar @values) {
				die;
			}

			# gather complete SQL query
			my $sql = ($t->{replace} ? 'REPLACE' : 'INSERT').
				' INTO '.$t->{name}.' ('.join(',',@cols).
				') VALUES ('.join(',',@values).')';

			# debug mode
			sw_log_debug( "ABSTRACT SAVE: $sql") if $DEBUG;

			$obj->{_webdb_}->select_run($sql);

			# if there was AUTO-property
			if ($prop_auto) {
				# update AUTO property (permanent)
				$vals->{$prop_auto} = $obj->{_webdb_}->select_inserted_id();
				# update related properties 
				# for direct-joined tables (local)
				foreach (grep { exists $props{$_}->{rel_prop} && 
					$props{$_}->{rel_prop} eq $prop_auto } keys %props) {
						$auto_vals{$_} = $vals->{$prop_auto};
				}
			}
		} else {
			# UPDATE EXISTING RECORD #

		    my @where=();
		    my @set=();

			## to shift object versions.

		    foreach my $prop_name (@prop_names){    	
		    	$prop = $props{$prop_name};

				if ($prop->{type} eq 'version') {
					## add to sql condition comparison with loaded version.
					push (@where, $prop->{col}.'='.$vals->{$prop_name});

					my $new_ver = $vals->{$prop_name}+1;
					## add to sql "update" command increment of object version.
					push(@set,$prop->{col}.'='.$new_ver);

					## mark version property to increment in "in-memory" object after successful save.
					$change_props{$prop_name} = $new_ver;
				} else {
					if($prop->{key}) {
						push(@where,$prop->{col}.'='.HSPC::WebDB::quote(exists $auto_vals{$prop_name} ? 
						$auto_vals{$prop_name} : $vals->{$prop_name}));
					} else {
						# by default we use quoted attribute value
						my $value = "''";
						if (	exists $prop->{type} && $prop->{type} eq 'lang' ) {
							$value = HSPC::WebDB::quote($vals->{$prop_name}->to_db_format()) if (
								ref($vals->{$prop_name})
							);
						} else {
							$value = HSPC::WebDB::quote($vals->{$prop_name});
						}
						# translate value if needed
						$value = _translate_case($value,{ reverse %{$prop->{translate}} })
							if exists $prop->{type} && $prop->{type} eq 'translate';
						# add assignment into SET list
						push(@set,$prop->{col}.'='.$value);
					}    	    	
				}
		    }		    
	        
			# gather complete SQL query
	    	my $sql = 'UPDATE '.$t->{name}.' SET '.join(',',@set).
	    		' WHERE '.join(' AND ',@where);

			# debug mode
			&sw_log_debug("ABSTRACT SAVE: $sql") if $DEBUG;

		    my $res = $obj->{_webdb_}->select_run($sql);
			sw_die("Object not updated. Object deleted, or object version is wrong, etc...")
				unless $res;

			## shift object versions in "in-memory" object after successful save.
		}
		while (my($prop_name, $new_val) = each(%change_props)) {
			$obj->{$prop_name} = $new_val;
		}
	}

	return;
}

############################################################################
# => id_col_name => $id # item id value
# < undef || errors_string # string with errors messages
sub delete {
    my $self = shift;
    my %h = (@_);

	## check - may be we try to delete object, which we load as read_only?
	sw_die("try to save object, which was load as 'read_only'")
		if ($self->{__object_was_load_as_read_only});

    # CAN BE DELETED ? #
	my $errors = $self->check_delete();
    return $errors if $errors;

	# turn debug mode on/off
	local $DEBUG = $h{debug} || 0;
    
    # DELETE #
    my $ci = $self->_get_class_info;
    my $props = $ci->{props};

	# start transaction
	my $trans = !$ci->{no_trans} ? trans(comment=>"Abstract->delete", webdb=>$self->{_webdb_}) : undef;

    # delete main object
   	foreach (keys %$self) {
   		$h{$_} = $self->{$_} unless exists $h{$_};
   	}

	_delete($self,$ci,\%h);

	# handle list properties
	foreach my $prop_name ( grep exists $props->{$_}->{type} && 
		$props->{$_}->{type} eq 'list' &&
		!$props->{$_}->{not_delete} && !$props->{$_}->{read_only}, sort keys %$props) {
		# delete list attribute
		_operate_list($self,$prop_name,'delete');
	}

	$trans->commit if !$ci->{no_trans};

    return;
}

################################ PRIVATE ###################################
# > CLASS_INFO
# > ref to hash with attribute values
sub _delete {
	my $obj = shift;
    my $ci = shift;
    my $vals = shift;
    my $list = shift || 0;

	my %tables = %{$ci->{tables}};
	my %props = %{$ci->{props}};

	# get name of the the main table
	my ($tmain_alias) = grep !exists $tables{$_}->{rel_table}, keys %tables;
	unless ($tmain_alias) {
		die q[Main table was not found for class];
	}

	# return if table marked as 'not for deletion'
	return if $ci->{tables}->{$tmain_alias}->{not_delete} ||
		$ci->{tables}->{$tmain_alias}->{read_only};

	# get list of all direct-joined tables marked for deletion
	my @joined_aliases = grep exists $tables{$_}->{rel_table} &&
		$tables{$_}->{rel_table} eq $tmain_alias && 
		$tables{$_}->{join_where} && $tables{$_}->{delete}, sort keys %tables;

	# add new properties for hard-joined tables
	_extend_properties($ci,\%props,$vals) if scalar @joined_aliases;

	my @sorted_joined_aliases;
	if ( scalar @joined_aliases ) {
		my %levels;
		if ( exists $ci->{level} && defined $ci->{level} ) {
			%levels = %{$ci->{level}};
		} else {
			sw_die("Script must not be here. This code is for foreign keys implementation. Ask billing team.");
		}
		my @sorted_tables = sort { $levels{$a} <=> $levels{$b} } keys %levels;

		foreach my $table (@sorted_tables) {
			if ( scalar grep { $table eq $_ } @joined_aliases ) {
				push(@sorted_joined_aliases, $table);
			}
		}
	}

	# update each table
	foreach my $t_alias ($tmain_alias,@joined_aliases) {
		# table properties
		my $t = $tables{$t_alias};
		# table real name
		my $t_name = $t->{name};

		# key attributes
		my @keyprop_names = grep exists $props{$_}->{table} &&
			$props{$_}->{table} eq $t_alias &&
			$props{$_}->{key}, keys %props;

		# die if all key attribs not defined
		unless (scalar (grep $vals->{$_}, @keyprop_names) > 0) {
			sw_die("ERROR: cannot delete record with all key attributes undefined");
		}

		# init vars #
		my @where=();

		# process key props #
		foreach my $prop_name (@keyprop_names) {
			unless (defined $vals->{$prop_name}) {
				if($list) {
					next;
				} else {
					die qq{ERROR: cannot delete record with key attribute $prop_name undefined};
				}
			}

			my $prop = $props{$prop_name};
			push(@where, $prop->{col}.'='.HSPC::WebDB::quote($vals->{$prop_name}));
		}

		my $sql = qq{DELETE FROM $t_name WHERE }.join(' AND ',@where);

		# debug mode
		sw_log_debug( "ABSTRACT DELETE: $sql") if $DEBUG;

		$obj->{_webdb_}->select_run($sql);
	}

    return;
}

############################################################################
# this function adds new properties for hard-joined tables (via join_where)
# to be used for separate processing of each table
#
# > reference to CLASS_INFO structure (used in read-only mode)
# > reference to properties hash (will be extended with new properties)
sub _extend_properties {
	my $ci = shift;
	my $xprops = shift;
	my $xvals = shift;

	my $tables = $ci->{tables};
	my $props = $ci->{props};

	# get name of the the main table
	my ($tmain_alias) = grep !exists $tables->{$_}->{rel_table}, keys %$tables;
	return unless $tmain_alias;

	# get list of all direct-joined tables marked for deletion
	my @joined_aliases = grep exists $tables->{$_}->{rel_table} &&
		$tables->{$_}->{rel_table} eq $tmain_alias && 
		$tables->{$_}->{join_where}, keys %$tables;

	# register new properties for hard-joined tables
	foreach my $t_alias (@joined_aliases) {
		foreach (split(/\s+AND\s+/,$tables->{$t_alias}->{join_where})) {
			# parse where condition
			next unless /(\w+)\.(\w+)\s*=\s*(\w+)\.(\w+)/;
			my ($t1,$v1,$t2,$v2) = $1 eq $tmain_alias ? 
				($1,$2,$3,$4) : ($3,$4,$1,$2);
			# $t1 should be $tmain_alias
			next unless $t1 eq $tmain_alias;

			# get main property name
			my ($main_prop) = grep { exists $props->{$_}->{col} &&
				$props->{$_}->{col} eq $v1 } keys %$props;
			next unless $main_prop;

			# check if property already exists
			my ($exist_prop) = grep { exists $props->{$_}->{col} && 
				$props->{$_}->{col} eq $v2 && $props->{$_}->{table} eq $t2 } keys %$props;

			if($exist_prop) {
				$xprops->{$exist_prop}->{key} = $props->{$main_prop}->{key};
				$xprops->{$exist_prop}->{rel_prop} = $main_prop;
			} else {
				# register new property
				my $new_prop = '_'.$t2.'_'.$v2;
				$xprops->{$new_prop} = { 'key'=>1,	%{$props->{$main_prop}},
					'col'=>$v2, 'table'=>$t2, 'rel_prop'=>$main_prop};
				$xvals->{$new_prop} = $xvals->{$main_prop};
			}
		}
	}
	return;
}

############################################################################
# > object reference
# < undef || errors_string # string with errors messages
sub check{
    my $self = shift;
    return;
}

############################################################################
# > object reference
# < undef || errors_string # string with errors messages
sub check_delete{
    my $self = shift;
    return;
}

############################################################################
# > class
# >= request # REQ_Y #
# => gate_cols
# >= where
# >= join
# => select_for_update
# >= tables
# >= debug
# >= lang
sub load_data{
	my $class = shift;
	my %h=(@_);
	my $req = $h{request};		
	my $lang = $h{lang} || current_lang();
	my $ci = _get_class_info($class);
	my $webdb = new HSPC::WebDB(dbname => $ci->{database});

	my $gate_cols=$h{gate_cols};		
	my $tables = $h{tables};
	my $join = $h{join};
	my $where = $h{where};

	my $prop;
	my $prop_name;
	my $t;
	my $i;
	my $used_tables;

	sw_die("try to load object, which not designed to use database")
		if $ci->{no_database};

	# find main table #
	foreach my $t_alias(keys %{ $ci->{tables} } ){
		$t = $ci->{tables}->{$t_alias};		
		next if $t->{rel_table};
		$used_tables->{$t_alias} = 1;
		last;
	}
	die unless $used_tables;

	### process cols ####
	foreach my $col (@{ $req->{cols} }){
		$prop_name = $col->{code};
		$prop = $ci->{props}->{$prop_name};
		if($prop){
			$col->{type} = $prop->{type} unless defined($col->{type});
			my $column;
			if($prop->{type} && $prop->{type} eq 'translate') {
				# if column value should be translated into attribute value
				$column = _translate_case(
					$prop->{table}.'.'.$prop->{col},$prop->{translate}
				);
			} elsif ( $prop->{type} && $prop->{type} eq 'lang') {
				$column = lang_field_select_expr($prop->{table}.'.'.$prop->{col},$lang);
			} else {
				if(exists $prop->{sql}) {
					# if column expressed via custom SQL code
					$column = $prop->{sql};
				} elsif(exists $prop->{col}) {
					# ordinary column definition (table + column)
					$column = $prop->{table}.'.'.$prop->{col};
				}
			}
			# expand gate_cols #
			$gate_cols->{$prop_name} = $column;

			if(defined $prop->{table}) {
			
				# mark table as used #			
				$used_tables->{$prop->{table}} = 1;
			
				# check related table;
				$t = $prop->{table};

				for($i=0;$i<=100;$i++){
					$t = $ci->{tables}->{$t};
					last unless defined $t->{rel_table};
					$used_tables->{$t->{rel_table}} = 1;
					$t = $t->{rel_table};
				}
				die qq[Too many joins to $prop->{table}] if $i==100;
			}
		}else{
			$gate_cols->{$prop_name} = $gate_cols->{$prop_name} || 'NULL';
			#die qq[ERROR: no prop with name "$prop_name"];
		}
	}			
	###	process used tables #
	foreach my $t_alias (sort { $ci->{tables}->{$a}->{order} <=> $ci->{tables}->{$b}->{order} } keys %$used_tables) {
		$t = $ci->{tables}->{$t_alias};		
		next unless $t; # may be this table defined as JOIN ?
		my $name = $t->{name}.' '.$t_alias;
		
		if( $t->{join_where} ){									
			$tables.=',' if $tables;
			$tables.=qq{$name};
			$where.=q{ AND } if $where;
			$where.=$t->{join_where};
		}elsif( $t->{join_inner} ){
			die unless $t->{rel_table};
			$join.=qq{ INNER JOIN $name ON $t->{join_inner} };
		}elsif( $t->{join_left} ){
			die unless $t->{rel_table};
			$join.=qq{ LEFT JOIN $name ON $t->{join_left} };
		}elsif( $t->{join_right} ){
			die unless $t->{rel_table};
			$join.=qq{ RIGHT JOIN $name ON $t->{join_right} };
		}else{
			die if $t->{rel_table};
			$tables.=',' if $tables;
			$tables.=qq{$name};
		}		
	}

	$h{gate_cols} = $gate_cols;
	$h{tables} = $tables;
	$h{join} = $join;
	$h{where} = $where;
	
	$h{debug} = 1 if $ENV{HSPC_DEBUG};

	return $webdb->select_data(%h);
}


## all params as for &_list, excluding select_count_only
## all params as for &_list, excluding select_count_only
## warining: you should select attributes for &count from
##	all tables to correct work
##	i.e. if there will be only attributes from parent
##	table in case of one-to-many join, joining will not
##	be actually done!
sub count
{
	return _list(@_, select_count_only => 1);
}

## all params as for &_list, excluding select_count_only
## all params as for &_list, excluding select_count_only
## warining: you should select attributes for &count from
##	all tables to correct work
##	i.e. if there will be only attributes from parent
##	table in case of one-to-many join, joining will not
##	be actually done!
sub list 
{
	return _list(@_, select_count_only => undef);
}

############################################################################
# list all items conformed to filtering settings and ordered appropriately
#
# > class
# >= attributes - attributes to be mentioned in output (reference to ARRAY)
# >= filter - filtering settings (reference to HASH)
# >= distinct - a boolean, its value is passed to the load_data call (true 
#		by default)
# >= select_count_only - as it means
#
# Not supported temporarily:
# >= sorting - reference to array of hashes with sorting settings
sub _list {
	my $class = shift 
		|| sw_die("ERROR: Class method called as library function");
	$class = ref $class if ref $class;
	my %h=(distinct	=> 1, @_);

	# turn debug mode on/off
	local $DEBUG = $h{debug} || 0;

	# get CLASS_INFO
	my $ci = _get_class_info($class);

    # init vars #
    $h{filter} ||= {};
    $h{attributes} ||= [ grep $ci->{props}->{$_}->{key}, keys %{$ci->{props}} ];

	## build DataRequest object
	my @cols = map { {code=>$_} } @{$h{attributes}};
	my @where_add = ();
	## transform filter into DataRequest + additional "where" conditions
	foreach my $attr (keys %{$h{filter}}) {
		my $prop = $ci->{props}->{$attr};
		my $col = $prop->{type} && $prop->{type} eq 'translate' ? 
			_translate_case($prop->{table}.'.'.$prop->{col}, $prop->{translate}) :
				$prop->{sql} || $prop->{table}.'.'.$prop->{col};

		## check if filter contain NULL value
		if(!defined $h{filter}->{$attr}) { ## scalar context
			push @where_add, "$col IS NULL";
			push @cols, { code=>$attr }; ## do not forget to select the column

		} elsif(ref $h{filter}->{$attr} eq 'ARRAY' &&
				scalar grep { !defined $_ } @{$h{filter}->{$attr}}) { ## array context
			my @vals = grep { defined $_ } @{$h{filter}->{$attr}};
			my $condition = scalar @vals > 1 ? "$col IN (".
				join(',',map {HSPC::WebDB::quote($_)} @{$h{filter}->{$attr}}).')' :
					scalar @vals == 1 ? "$col = ".HSPC::WebDB::quote($vals[0]) : undef;
			$condition = $condition ? "$col IS NULL OR $condition" : "$col IS NULL";
			push @where_add, $condition;
			push @cols, { code=>$attr }; ## do not forget to select the column

		## check if we are filtering by empty array
		} elsif(ref $h{filter}->{$attr} eq 'ARRAY' && !scalar @{$h{filter}->{$attr}}) {
			return $h{select_count_only} ? 0 : [];

		} else {
			my $val = $h{filter}->{$attr};
			push @where_add, ref $h{filter}->{$attr} eq 'ARRAY' ?
				"$col IN (".join(',',map {HSPC::WebDB::quote($_)} @{$val}).')' :
					"$col = ".HSPC::WebDB::quote($val);
			push @cols, { code=>$attr };
		}
	}
	my $request = HSPC::DataRequest->new(cols => \@cols);

	## add newly constracted "where" conditions
	$h{where} .= ($h{where} ? ' AND ' : '').'('.join(') AND (',@where_add).')' 
		if scalar @where_add;

	## load data from DB
	&load_data($class,request=>$request, distinct=>$h{'distinct'}, debug=>1,
					  map { $_ => $h{$_} } qw(gate_cols tables join where select_count_only));

	return $request->{total_items} if $h{select_count_only};

	## drop extra data
	foreach (@{$request->{data}}) {
		$#{$_} = $#{$h{attributes}};
	}

	return $request->{data};
}

############################################################################
## > class
## < class_tiltle_id - string_id for class title
sub class_title_id
{
	my $class = shift;
	my $self;
	if ( ref($class) ) {
		## constructed object
	 $self = $class;
	} else {
		$self = new($class,@_);
	}

	## return title_id if we has it already
	return $self->_get_class_info->{class_title_id} if (
		exists($self->_get_class_info->{class_title_id})
	);

	## find title_id in table
	my $type_id = $self->obj_type_id();
	my $result = select_result(
		"select name_id from core_node_type where type_id = ?",$type_id
	);
	$self->_get_class_info->{class_title_id} = $result;

	return $self->_get_class_info->{class_title_id};
}## /class_title_id

############################################################################
## > class
## < object type - id for class type.
sub obj_type_id
{
	my $class = shift;
	my $obj_type_id = ref($class)||$class;
	return $obj_type_id;
}## /obj_type_id


## clone object
## Argument description see in _clone and _erase_keys methods
sub clone {
	my $self = shift;
	my %h = (
		erase_auto_keys       => 1,
		erase_class_info_keys => 1,
		load_lists            => 1,
		erase_all_keys        => undef,
		load_lists_force      => undef,
		@_
	);

	unless(ref($self) && $self->isa(__PACKAGE__)) {
		sw_die(__PACKAGE__."::clone is object method");
	}

	## get class to avoid any methods to be called on self
	my $class = ref($self);
	
	## collect _class_info_ and _web_db_
	## Create new object skipping childs new methods
	my $new_obj = new($class);

	## fill clone
	$self->_clone(
		new_obj => $new_obj,
		load_lists => $h{load_lists},
		load_lists_force => $h{load_lists_force},
	);

	## delete keys from clone object
	$new_obj->_erase_keys(
		erase_auto_keys       => $h{erase_auto_keys}       || '',
		erase_class_info_keys => $h{erase_class_info_keys} || '',
		erase_all_keys        => $h{erase_all_keys}        || '',
	);


	return $new_obj;	
}


## Copy all data
## >> self - to be cloned
## =>> new_obj - clone
## => load_lists - load data of list attributes from db if source object did not load them 
## => load_lists_force - load data of list attributes from db
sub _clone {
	my $self = shift;
	my %h = (
		new_obj => undef,
		load_lists=>undef,
		load_lists_force=>undef,
		@_
	);
	my $new_obj = $h{new_obj};
	my $load_lists = $h{load_lists} || $h{load_lists_force};
	my $load_lists_force = $h{load_lists_force};
	unless(
		ref($new_obj) && $new_obj->isa(__PACKAGE__) &&
		ref($new_obj) eq ref($self)
	) {
		sw_die(__PACKAGE__."::_clone \$h{new_obj} is incorrect");
	}
	
	my $ci = $self->_get_class_info;
	my $props = $ci->{props};
	
	## find all 'simple' attributes and copy their values
	my @prop_names = grep { $props->{$_}->{type} ne 'list' } keys %$props;
	@{$new_obj}{@prop_names} = @{$self}{@prop_names};
	
	## copy custom keys
	my @custom = ('__object_was_load_as_read_only','__object_was_locked',);
	@{$new_obj}{@custom} = @{$self}{@custom};
	
	## for debuging store clone flag
	$new_obj->{__clone_copy}++;
	
	## copy list attributes
	
	for my $prop_name (grep { $props->{$_}->{type} eq 'list' } keys %$props) {
		
		my $prop_list = $self->_get_class_info->{props}->{$prop_name};
		my $subci = $prop_list->{_class_info_};
		my $prop_class = $prop_list->{class_ref};
		$subci or die qq{ERROR: no class_info found for $prop_name attribute};
		my $subprops = $subci->{props};
		
		## if data where not loaded or force 
		## load data from db for this attribute 
		if(
			(!defined $self->{$prop_name} && $load_lists) ||
			$load_lists_force
		) {
			## find key attributes
			my @defined_keys = grep { $props->{$_}->{key} && $new_obj->{$_} } keys %$props;
			$new_obj->$prop_name if scalar @defined_keys; ## force load list
			$new_obj->{$prop_name} = [] unless defined $new_obj->{$prop_name};
		}
		## copy from object
		elsif(ref $self->{$prop_name}) {
			## list attribute should not be in class_info of list attribute
			my @prop_names = grep { $subprops->{$_}->{type} ne 'list' } keys %$subprops;
			
			$new_obj->{$prop_name} = [];
			for (grep {ref $_} @{$self->{$prop_name}}) {
				my $list_copy = $prop_class ? $prop_class->new : {};
				@{$list_copy}{@prop_names} = @{$_}{@prop_names};
				push(@{$new_obj->{$prop_name}},$list_copy);
			}
		}
		else {
			## it is strange if we are here 
			$new_obj->{$prop_name} = $self->{$prop_name};
		}
	}
	
}

## Desc: unset key attributes
## >> self
## => erase_auto_keys - set to undef values of all keys marked as 'AUTO'
##     (including attributes of list attributes)
## => erase_class_info_keys - set to undef values of 'main' keys
##     (including corresponding attributes of list attributes)
## => erase_all_keys - set to undef values of all keys
##     (including attributes of list attributes)
sub _erase_keys {
	my $self = shift;
	my %h = (
		erase_auto_keys       => undef,
		erase_class_info_keys => undef,
		erase_all_keys        => undef,
		@_
	);
	my $ci = $self->_get_class_info;
	my $props = $ci->{props};
	unless(grep { $h{$_} } keys %h ) { 
		## nothing to be deleted
		return; 
	}
	
	## find all 'key' attributes in class_info
	my @ci_keys2erase = ();
	if($h{erase_all_keys} || $h{erase_class_info_keys}) {
		@ci_keys2erase = grep { $props->{$_}->{key} } keys %$props;
		
	}
	elsif($h{erase_auto_keys}) {
		@ci_keys2erase = grep { $props->{$_}->{key} eq 'AUTO' } keys %$props;
	}
	
	## delete values
	@{$self}{@ci_keys2erase} = ();
	
	########################################
	## delete keys from list attributes

	my %ci_keys2erase;
	for (grep { $props->{$_}->{key} } keys %$props) {
		$ci_keys2erase{$_}=1;
	}
		
	for my $prop_name (
		grep { 
			$props->{$_}->{type} eq 'list' &&  ref($self->{$_})
		} keys %$props
	) {
		## find class info of list attribute
		my $prop_list = $self->_get_class_info->{props}->{$prop_name};
		my $subci = $prop_list->{_class_info_};
		$subci or die qq{ERROR: no class_info found for $prop_name attribute};
		my $subprops = $subci->{props};
		
		## list attribute should not be in class_info of list attribute
		my %keys2erase = ();
			
		if($h{erase_auto_keys}) {
			for (grep { $subprops->{$_}->{key} eq 'AUTO' } keys %$subprops) {
				$keys2erase{$_} = 1;
			}
		}
			
		if($h{erase_all_keys}) {
			for (grep { $subprops->{$_}->{key} } keys %$subprops) {
				$keys2erase{$_} = 1;
			}
		}
		
		if($h{erase_class_info_keys}) {
			## delete corresponding 'main' keys
			my %join = %{$prop_list->{join}};
			for (grep { $ci_keys2erase{$join{$_}} } keys %join) {
				$keys2erase{$_} = 1;
			}
		}
				
		if(keys %keys2erase) {
			for (grep {ref $_} @{$self->{$prop_name}}) {
				@{$_}{keys %keys2erase} = ();
			}
		}
	}
}

## cc_view_url -- get url for view object
## >> self
## < url
sub cc_view_url
{
	my $self = shift;
	sw_die('HSPC::MT::Abstract->cc_view_url -- invalid param')
		if(@_);
	
	## Get screen alias
	my $url = HSPC::MT::Core::ObjType->list(
		attributes => ['alias','uri_params'],
		filter => { type_id => $self->obj_type_id() }
	);

	return '' unless ( @{$url} && $url->[0]->[0] );
	
	my $node = $ENV{COMPREP}->node(alias => $url->[0]->[0]);
	## Test security
	return unless ( $node && $ENV{security_obj}->test_security($node->{alias}) );
	## Link
	return $ENV{COMPREP}->link(node_id => $node->{node_id}).'/'.$url->[0]->[1];
}


## return 1 if obj has been changed
sub is_mt_obj_changed {
	my $self = shift;
	$self->{_is_mt_obj_changed} = shift if ( @_ );
	return $self->{_is_mt_obj_changed};
}##/is_mt_obj_changed

1;
__END__

HSPC::MT::Abstract - base class for MT objects

