package HSPC::MenuTree;

use strict;
use XML::Parser;
use HSPC::Localization;
use Data::Dumper;
use SWsoft::protect;	# Mark this module for encryption upon build

my ($menu_el, $self, @stack);
my $debug = 0;
my @tt = qw|caption level closed_icon url open_icon id security frame|;

sub new {
	my ($classname, $path) = @_;

	$path =~ s/[\0]//g; ## remove char with ASCII code equal 0
	$path =~ s/[\s\t\&\;\`\'\\\"\|\*\?\~\<\>\^\(\)\[\]\{\}\$\n\r]//g; ## remove special chars
	$path =~ s/\.\.//g; ## remove .. from path

	unless ( -e $path && -s $path && -r $path && ! -x $path) {
		return die "Can't open $path!";
	}

	undef $/;
	sysopen FH, $path, "O_RDONLY" or die "Can't open $path! $@";
	my $xml = <FH>;
	close(FH);

	my $p = new XML::Parser;
	$p->setHandlers(
		Init => \&_init,
		Start => \&_start,
		End => \&_end,
	);	
	$p->parse($xml);	
	undef $p;
	_post_process($self);
	warn 'DEBUG: HSPC::MenuTree->new returned ' . Dumper($self) if $debug;
	return bless $self, $classname;
}

sub traverse {
	my ($node, $security, $func, $id_func) = @_;
	
	return unless _is_visible($node, $security);
		

	&$id_func($node) if $node->{id} ne '';

	# prepare data #		
	my @items = map{$node->{$_}} @tt;
	
	for (my $i = $#items; $i >= 0; $i--) {
		last unless $items[$i] eq '';
		pop @items;
	}
	$node->{data} = join(', ',@items);	
	# /prepare data #

	&$func($node);

	my $children = $node->{children};
	return unless $children;
	foreach (@$children) {
		traverse($_, $security, $func,$id_func);
	}
}

sub _init {
  $menu_el = 0;
}

sub _start {
  my $cname = shift;
  my $tag = shift;
  	    	    	  
	if ($tag eq 'menu') {
		die 'Only one "menu" element allowed, and it must be document root element.' if $menu_el || $#stack != -1;
		$self = _gen_tree_item(@_);
		push @stack, $self;
		$menu_el = 1;		
		return;
	}
  
	die 'No other elements except "menu" and "node" permitted.' unless $tag eq 'node';
	die '"node" outside "menu" element is not permitted.' unless $menu_el;
	
	my $children = $stack[-1]->{children};
	my $rec = _gen_tree_item(@_);
	push @$children, $rec;
	push @stack, $rec;
}

sub _end {
	my ($cname, $tag) = @_;
	$menu_el = 0 if $tag eq 'menu';
	my $rec = pop @stack;
	my $children = $rec->{children};
	delete $rec->{children} if $#$children == -1;
}

sub _gen_tree_item {
	my %args = @_;			
	
	my $ret = {
		caption => $args{dynamic} ? 'Loading...' : ($args{caption}?string_lang('en',$args{caption}):''),
		level => $#stack + 1,
		url => $args{url} ? $args{url} : '',
		open_icon => $args{'open'} ? $args{'open'}:'',
		closed_icon => $args{'closed'} ? $args{'closed'} : '',
		id => $args{id} ? $args{id} : '',
		security => $args{dynamic} ? '0' : '0',
		frame => $args{frame} ? $args{frame} : '',
						                
	    children=> []
	};
            
	$ret->{sc_id} = $args{sc_id} if defined $args{sc_id};
	$ret->{acct_type} = $args{acct_type} if defined $args{acct_type};

	return $ret;
}

sub _post_process {
	my ($node) = @_;			
	
	# HANDLE ICONS #
	if($node->{closed_icon} ne '' &&  !($node->{closed_icon}=~m#\.#)){
		$node->{closed_icon} = $ENV{HSPC_SYSTEM}->ui_image_path(id=>$node->{closed_icon});        
	}
	#$node->{open_icon} = $node->{closed_icon} if $node->{open_icon} eq '';
	# /HANDLE ICONS #			
		
    if ( defined $node->{acct_type} ) {
		$node->{acct_type} = { map {$_, 1} split /\s*,\s*/, $node->{acct_type} };
    }

    if ( defined $node->{sc_id} && $node->{sc_id} =~ /\s*,\s*/ ) {
           $node->{sc_id} = [ split /\s*,\s*/, $node->{sc_id} ];
    }

	my  $children = $node->{children};
	return unless $children;

    foreach (@$children) {
		_post_process($_);
    }
}

sub _test_security {
	my ($node, $security) = @_;

	my $sc_id = $node->{sc_id};

	defined $sc_id or do {
		return _has_any_visible_child(@_);
	};

	ref($sc_id) eq 'ARRAY' or do {
		return 1 if $sc_id eq '';
		return $security->test_security($sc_id);
	};

	foreach (@$sc_id) {
		return 1 if $security->test_security($_);
	}
	return 0;
}

sub _is_visible {
	my ($node, $security) = @_;

        _test_security(@_) or do {
		return 0;
	};

        if (defined $node->{acct_type}) {
		$node->{acct_type}->{$security->account_type} or do {
			return 0;
		};
        }
	return 1;
}

sub _has_any_visible_child {
	my ($node, $security) = @_;
	
	my $children = $node->{children};
	return 0 unless $children;
	foreach (@$children) {
		return 1 if _is_visible($_, $security);
	}
	return 0;
}

1;

__END__
