# DB functions
# Author: baza@

use strict;
package HSPC::WebDB;

use Exporter;
use DBI;
use Error qw(:try);
use HSPC::Exc::Fatal;
use HSPC::Exc::Mysql;

use vars qw(@ISA @EXPORT @EXPORT_OK);
use HSPC::MT::Core::Constants;
use SWsoft::protect;	# Mark this module for encryption upon build

use HSPC::Application ();
use HSPC::Logger;

our $VERSION = 1.0;

@ISA = qw(Exporter);
@EXPORT = qw(
  &select_run
  &select_rows
  &select_data
  &select_hash  
  &select_result
  &select_commit
  &select_rollback
  &select_row
  &select_column
  &select_inserted_id  
  &update_obj
  &insert_obj
  &load_obj
  &select_hashrows
  &select_open
  &select_commit
  &quote
  &select_hash_all  
  &delete_obj
  &is_table_exists
  &is_column_exists
  &find_column
  &find_keys
  &is_key_exists
);

@EXPORT_OK = qw(
&mutex &amutex
&trans &atrans
&trans_counter
);
################################################################################
## Usage:  my $mutex = mutex(comment=>"Processing Order");
## Description: create new mutex.
## Parameters:
##     mutex   =>> mutex
##     comment => comment
##     timeout => timeout (seconds). If lock not obtained during timeout - die. May be zero to poll.
##     sleep_time => time in milliseconds to sleep between tries.
## Return value:
##		HSPC::Mutex object.
sub mutex {
	my %h = (
		mutex      => undef,
		comment    => undef,
		timeout    => undef,
		sleep_time => undef,
		@_
	);

	require HSPC::Mutex;
	my $mutex = HSPC::Mutex->new(
		mutex      => $h{mutex},
		comment    => $h{comment},
		timeout    => $h{timeout},
		sleep_time => $h{sleep_time},
	);

	return $mutex;
}

################################################################################
## Usage:  my $mutex = amutex(comment=>"Processing Order");
## Description: create new mutex.
## Parameters:
##     mutex   =>> mutex
##     comment => comment
##     timeout => timeout (seconds). If lock not obtained during timeout - die. May be zero to poll.
##     sleep_time => time in milliseconds to sleep between tries.
## Return value:
##		HSPC::Mutex object.
sub amutex {
	my %h = (
		mutex      => undef,
		comment    => undef,
		timeout    => undef,
		sleep_time => undef,
		@_
	);

	my (undef, $filename, $line, $subr) = caller(1);

	$subr =~ s/(.*)::/$1->/;

	return mutex(
		mutex      => $h{mutex},
		comment    => "$subr $h{comment}",
		timeout    => $h{timeout},
		sleep_time => $h{sleep_time},
	);
}


################################################################################
## Usage:  my $trans = trans(comment=>"Saving Order");
## Description: create new transaction.
## Parameters:
##		comment => comment
##		dbname => database name, MAIN by default.
## Return value:
##		HSPC::Transaction object.
sub trans {
	my %h = (
		comment	=> undef,
		dbname	=> undef,
		@_
	);
	require HSPC::Transaction;
	my $trans = HSPC::Transaction->new(
		comment	=> $h{comment},
		dbname	=> $h{dbname}
	);

	return $trans;
}

################################################################################
## Usage:  my $trans = atrans(comment=>"Saving Order");
## Description: create new transaction and autmatically make comment.
## Parameters:
##		comment => comment
##		dbname => database name, MAIN by default.
## Return value:
##		HSPC::Transaction object.

sub atrans {
	my %h = (
		comment	=> undef,
		dbname	=> undef,
		@_
	);

	my (undef, $filename, $line, $subr) = caller(1);

	$subr =~ s/(.*)::/$1->/;

	return trans(comment=>"$subr $h{comment}", dbname=>$h{dbname});
}

## dbname => database name or undef for MAIN
## < count of open transactions
sub trans_counter{
	my %h = @_;

	$ENV{HSPC_DB_TRANS} = {} unless ref($ENV{HSPC_DB_TRANS});

	my $dbname = $h{dbname} || 'MAIN';
	return $ENV{HSPC_DB_TRANS}->{$dbname} || 0;
}


## Now class have only one attribute
## dbname - database name, by default = MAIN.
our %instances;
sub new {
	my $class = shift;
	my %h = (dbname => undef, @_);

	return $instances{$h{dbname}}
		if $instances{$h{dbname}};

	my $self = { dbname => $h{dbname} || 'MAIN' };
	bless $self, $class;

	return $instances{$h{dbname}} = $self;
}

## accessor for dbname
sub dbname { $_[0]->{dbname} }

sub db_connect {
	my $self = shift;
	return ref($ENV{system_obj})
		? $ENV{system_obj}->db_connection(dbname => $self->dbname)
		: HSPC::Application->db_connection(dbname => $self->dbname);
}

# > sql
sub trace_open{
	my $self = shift if (ref($_[0]));
	my $msg = shift;

	my $logger = HSPC::Logger->new(category => 'sql_log');
	local $Log::Log4perl::caller_depth += 1;
	$logger->debug($msg);
}

sub trace_close{
	my $self = ref($_[0]) ? shift : undef;
	return unless ref($ENV{system_obj});
}

# > sql
# > params
# < rows processed 
sub select_run{
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my $sql = shift;
	my @arr_params = @_;
	for(my $i=0; $i <= $#arr_params; $i++)	{ 
		$arr_params[$i]='' unless (defined $arr_params[$i]); 
	}	
	my $dbh = $self->db_connect;
	throw HSPC::Exc::Fatal( msg => "no dbh obtained" ) unless $dbh;
	my $sth = $dbh->prepare($sql);
	
	my $tmp_err = $@;

	$self->trace_open($sql);
	eval{		
		$sth->execute(@arr_params);			
	};
	my $err = $@;

	$self->trace_close();
	
	#### check errors
	if ($err) {
		$sth->finish();	
		throw HSPC::Exc::Mysql(
			mysql_errno => $dbh->{'errno'}, ## XXX
			msg         => "${err}${sql}"
		);
	}
		
	$sth->finish();	

	## restore eval error. It's nesessary, when executed on destructor of mutex inside other eval.
	$@ = $tmp_err;

	return $sth->rows;
}

## =>> webdb object
## < connection id
sub connection_id {
	my $self = shift;
	
	my $dbh = $self->db_connect;
	return $dbh->{'mysql_thread_id'};
};

# > sql
# > params
# < result
sub select_rows{
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my $sql = shift;
	my $dbh = $self->db_connect;				
	my $sth = $dbh->prepare($sql);	
	#arn $sql;
	my @arr_params = @_;
	for(my $i=0; $i <= $#arr_params; $i++)	{ 
		$arr_params[$i]='' unless (defined $arr_params[$i]);
	}					
	
	$self->trace_open($sql);
	eval{
		$sth->execute(@arr_params)		
	};
	my $err = $@;
	$self->trace_close();
	
	#### check errors
	if ($err) {
		$sth->finish();	
		throw HSPC::Exc::Mysql( msg => "${err}${sql}" );
    }
	
	my $rows = $sth->fetchall_arrayref();
	$sth->finish();
	return $rows;
}

# > sql
# > params
# < result
sub select_column{
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my $sql = shift;
	my $dbh = $self->db_connect();
	my $sth = $dbh->prepare($sql);
	my @arr_params = @_;
	for(my $i=0; $i <= $#arr_params; $i++)	{
		$arr_params[$i]='' unless (defined $arr_params[$i]);
	}	
	$self->trace_open($sql);
	eval{
		$sth->execute(@arr_params);
	};
	my $err = $@;
	$self->trace_close();
	
	#### check errors
	if ($err) {
		$sth->finish();	
		throw HSPC::Exc::Mysql( msg => "${err}${sql}" );
    }
	
	my @rows;
	while (my $tec_result=$sth->fetchrow_array) { 
		push(@rows,$tec_result);
	}
	$sth->finish();
	return \@rows;
}

## > into_hash  : not REQ
## > sql
## > params
## < in_hash defined : 0 for empty rows and 1 in another case
##    OR 
## < ref to array with �ols
sub select_row {
	my $self = ref($_[0]) eq 'HSPC::WebDB' ? shift : new HSPC::WebDB;
	my $into_hash = ref($_[0]) ne '' ? shift : undef;	
	my $sql = shift;

	my $dbh = $self->db_connect;			
	my $sth = $dbh->prepare($sql);		
		
	my @arr_params = @_;
	for(my $i=0; $i <= $#arr_params; $i++)	{
		$arr_params[$i]='' unless (defined $arr_params[$i]);
	}	
	$self->trace_open($sql);
	eval{
		$sth->execute(@arr_params);	
	};
	my $err = $@;
	$self->trace_close();
	
	#### check errors
	if ($err) {
		$sth->finish();	
		throw HSPC::Exc::Mysql( msg => "${err}${sql}" );
    }
		
	if( defined $into_hash ){
		my $h = $sth->fetchrow_hashref;
		return 0 unless $h;
		@$into_hash{keys %$h} = values %$h;
		return 1;
	}
	my @result = $sth->fetchrow_array;			
	$sth->finish();
	return \@result;
}


# > sql
# > params
# < hash
sub select_hash{
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my $sql = shift;

	my $dbh = $self->db_connect;		
	my $sth = $dbh->prepare($sql);		
			
	my @arr_params = @_;
	for(my $i=0; $i <= $#arr_params; $i++)	{
		$arr_params[$i]='' unless (defined $arr_params[$i]);
	}	
	
	$self->trace_open($sql);		
	eval{
		$sth->execute(@arr_params)		
	};
	my $err = $@;
	$self->trace_close();
	
	#### check errors
	if ($err) {
		$sth->finish();	
		throw HSPC::Exc::Mysql( msg => "${err}${sql}" );
    }
	
	my $h = $sth->fetchrow_hashref;
	$sth->finish();
	return $h;
}

# > sql
# > params
# < [hash,hash,...]
sub select_hashrows{
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my $sql = shift;

	my $dbh = $self->db_connect;	
		
	my $sth = $dbh->prepare($sql);	
	
	my @arr_params = @_;
	for(my $i=0; $i <= $#arr_params; $i++)	{
		$arr_params[$i]='' unless (defined $arr_params[$i]); 
	}
	
	$self->trace_open($sql);
	eval{
		$sth->execute(@arr_params);
	};
	my $err = $@;
	$self->trace_close();
	
	#### check errors
	if ($err) {
		$sth->finish();	
		throw HSPC::Exc::Mysql( msg => "${err}${sql}" );
    }
	
	my $rows;
	while( my $h = $sth->fetchrow_hashref ){
		my $hash;
		@$hash{keys %$h} = values %$h;
		push  @$rows,$hash;
	}
	$sth->finish();
	return $rows;		
}

# > sql
# > params
# < result
sub select_result{
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my $sql = shift;
	my $dbh = $self->db_connect;		
	my $sth = $dbh->prepare($sql);
	
	my @arr_params = @_;
	for(my $i=0; $i <= $#arr_params; $i++)	{
		$arr_params[$i]='' unless (defined $arr_params[$i]); 
	}
	
	$self->trace_open($sql);		
	eval{		
		$sth->execute(@arr_params);			
	};
	my $err = $@;
	$self->trace_close();

	#### check errors
	if ($err) {
		$sth->finish();	
		throw HSPC::Exc::Mysql( msg => "${err}${sql}" );
    }
	
	my ($result) = $sth->fetchrow_array;
	$sth->finish();
	return $result;
}

# => request : {} HSPC::DataRequest
#		=> sort_col_code
#		=> sort_revers
#		=> from : from 1
#		=> count
#		=> filters {code=>value,code1=>value1}
#		=> cols - look to WebPage->prepare_browse() API  /OR/  plan array of col_code, example : [qw(id name count)]
#		=> select_as_hashes : select data as hashes array # REQ_N #
#		=> select_as_col : select data one-dimmension array # REQ_N #
#		=> select_only_count : load only items count, do not load items data itself
#       => hash_key_code : return results as HASH with key value from column with req. code 
#		=> debug
#       => distinct
#		=> having_sql
#		=> where_count_sql
# => gate_cols  {code=>'db_col',...}
# => tables
# => where
# => order
# => group 
# => join
# => distinct
# => debug
#
#	details about col format look at HSPC::Page->prepare_browse()
#		
#
# <= put results in src_info
#		data=>
#		total_items=>
sub select_data{
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my %h=(@_);
	my $r=$h{request};	
	my $cols = $r->{cols};					
	
	if(!$h{request}){
		throw HSPC::Exc::Fatal( msg => "no request for select_data()" );
	}
	if(!$h{request}->{cols}){
		throw HSPC::Exc::Fatal( msg => "no cols for select_data()" );
	}
	
	my ($sql_t,$sql_c,$sql_w,$sql_l,$sql,$sql_o,@wcols,@hcols,$wcol,@counter_cols);
	my $dbh = $self->db_connect;
	
	# init tables #
	$sql_o	= $h{'order'};
	$sql_t	= $h{'tables'};
	my $db_col;
		
	# init columns and orders #
	my $col_code;
	my $sorted;
	my $loaded_index=0;
	foreach my $col(@{$r->{cols}}){
		if( $col->{no_select} ){
			$col_code = ref($col) eq 'HASH'?$col->{code}:$col;
			$db_col = '\'\'';
		}else{
			$col_code = ref($col) eq 'HASH'?$col->{code}:$col;
			$db_col = $h{gate_cols}->{$col_code};		
		}
				
		unless(defined $db_col){
			my @info;						
			foreach(keys %{$h{gate_cols}} ){
				push @info,'code='.$_;
			}
			throw HSPC::Exc::Fatal( 
				msg => 'ERROR: select_data() : no column "'
						.$col_code.'" in gate_list, tables '.$h{tables}.', gate:'
						.join(',',@info)
			);
		};
		
		$sql_c .= ',' if $sql_c;				
		$sql_c .=$db_col;
		
		push @counter_cols,$db_col if $db_col ne 'NULL';
			
		if($col_code eq $r->{sort_col_code}){			
			$sorted = 1;
			$sql_o.=',' if $sql_o;
			if($col->{type} eq 'ip'){
				$sql_o.= qq|inet_aton($db_col)|;
			}else{
				$sql_o.= $loaded_index+1;
			}
			$sql_o.=' DESC ' if($r->{sort_revers});
		}				
		
		$col->{loaded_index} = $loaded_index;
		$loaded_index++;
	}		
	
	# sorting by non-existed in cols column #
	if(!$sorted && $r->{sort_col_code}){		
		$sorted = 1;
		$sql_o.=',' if $sql_o;
		$sql_o.= $h{gate_cols}->{ $r->{sort_col_code} };
		$sql_o.=' DESC ' if($r->{sort_revers});		
	}
	
	if($sql_c eq '' && (not defined $r->{select_only_count})){
		throw HSPC::Exc::Fatal(
			msg => 'ERROR : no linked cols in select_data(), tables='
					.$h{tables}.' where='.$h{where}
		);
	}
	
	# init where #
	if($h{where}){
		$h{where} = " ( $h{where} ) " unless $h{where} =~ /^[\s()]*$/;
	}
		

	# process filters #
	##$sql_w = '';
	foreach my $col(@{$r->{cols}}){				
		next unless $col->{for_search} || $col->{for_filter} || $col->{for_where};		
		next unless $col->{for_where} || (defined($col->{value}));
		next if $col->{no_auto_filter};		
					
		$db_col = $h{gate_cols}->{$col->{code}};		
		next if (!$db_col || $db_col eq 'NULL');
		##$sql_w.=' AND ' if $sql_w;
		
		my $expr = defined $col->{for_where}?
			( $col->{for_where}==1?'=':$col->{for_where} )   : 
			'=';
		
		undef $wcol;
		if(defined($col->{value})){						
			if( ref($col->{value}) eq 'ARRAY' ){
				$wcol .= ' (';
				my @ranges;
				my @values = map {quote($_)} grep {
					## remove range values, storing them in an array
					/^\s*(\d+)\s*-\s*(\d+)\s*$/
						&& push(@ranges, [quote($1), quote($2)]) ? 0 : 1
				} @{$col->{value}};
				$wcol .= "$db_col IN (" . join(',', @values) . ')';
				$wcol .= join(' OR ', '', ## empty element to produce first "OR"
					map {"($db_col >= $_->[0] AND $db_col <= $_->[1])"} @ranges)
					if @ranges;
				if (grep {$_ eq 'NULL'} @values) {
					$wcol .= " OR $db_col IS NULL ";
				}
				$wcol .= ') ';
			}else{
		
				my $values = [$col->{value}];			
				$wcol.=' ( ';
				my $index=0;
								
				foreach my $value(@$values){							
					$wcol.=q{ OR } if $index>0;
					if($col->{type} eq 'id' || $col->{type} eq 'int' || $col->{type} eq ''){			
						$wcol .=  $db_col.$expr.$dbh->quote(int($value));
					}elsif($col->{type} eq 'money'){
						$wcol .=  $db_col.$expr.$dbh->quote($value);
					}elsif($col->{type} eq 'options'){
						$wcol .=  $db_col.$expr.$dbh->quote($value);
					}elsif($col->{type} eq 'date'){
						$wcol .=  $db_col.$expr.$dbh->quote($value);
					}elsif($col->{type} eq 'datetime'){
						$wcol .=  $db_col.$expr.$dbh->quote($value);						
					}elsif($col->{type} eq 'bool'){
						$wcol .=  $db_col.$expr.$dbh->quote($value);
					}elsif($col->{type} eq 'enum'){  
                                                $wcol .=  $db_col.$expr.$dbh->quote($value);
					}else{
						$value =~ s#\\#\\\\#g;
						#### apply sql-where filter
						if((not defined $col->{for_where}) || $col->{for_where}==1){
							$value=~s|\%|\\%|g;
							$wcol .= 'UPPER('.$db_col.") LIKE ". $dbh->quote("%".uc($value)."%");
						}else{
							$wcol .=  $db_col.$expr.$dbh->quote($value);
						}
					}		
					$index++;		
				}
				$wcol.=' ) ';				
			}
		}else{
			$wcol = $db_col.$expr;
			##$sql_w .= $db_col.$expr;			
		}
		## store col #
		if($col->{for_having} && $h{group}){
			push @hcols,$wcol;
		}else{
			push @wcols,$wcol;
		}
		## /store col ##
	}
	
	## process ranges ##
	foreach my $col(@{$r->{cols}}){
		next unless $col->{for_range};		
				
		$db_col = $h{gate_cols}->{$col->{code}};		
		next unless $db_col;
		$wcol=undef;
		
		if($col->{type} eq 'id' || $col->{type} eq 'int' || $col->{type} eq ''
			|| $col->{type} eq 'datetime'
			|| $col->{type} eq 'date')
		{
			if( defined $col->{value_from} && $col->{value_from} ne ''){
				$wcol = ' ( ';
				$wcol .= $db_col.'>'.($col->{for_range}->{from_exc}?'':'=').quote($col->{value_from});
			}
			if( defined $col->{value_to} && $col->{value_to} ne ''){
				if(defined $wcol){				
					$wcol .= ' AND ';
				}else{
					$wcol =' ( ';
				}
				$wcol .= $db_col.'<'.($col->{for_range}->{to_exc}?'':'=').quote($col->{value_to});
			}
		}else{
			throw HSPC::Exc::Fatal(
				msg => 'error: col_type='.$col->{type}	.', col_code='.$col->{code}
						.', can`t be uses in range=>'
			);
		}		
		
		next unless defined $wcol;
		
		## store col ##
		$wcol.=' ) ';
		if($col->{for_having} && $h{group}){
			push @hcols,$wcol;
		}else{
			push @wcols,$wcol;
		}
		## /store col ##
	}
	#
	
	
	## process frequency ##
	foreach my $col(@{$r->{cols}}){
		next unless $col->{for_frequency};		
				
		$db_col = $h{gate_cols}->{$col->{code}};		
		next unless $db_col;
		$wcol=undef;
		
		if($col->{type} eq 'datetime' || $col->{type} eq 'date'){
			if( defined $col->{value_from} && $col->{value_from} ne ''){
				$wcol = ' ( ';
				$wcol .= $db_col.'>'.($col->{for_frequency}->{from_exc}?'':'=').quote($col->{value_from});
			}
			if( defined $col->{value_to} && $col->{value_to} ne ''){
				if(defined $wcol){				
					$wcol .= ' AND ';
				}else{
					$wcol =' ( ';
				}
				$wcol .= $db_col.'<'.($col->{for_frequency}->{to_exc}?'':'=').quote($col->{value_to});
			}
		}else{
			throw HSPC::Exc::Fatal(
				msg=> 'error: col_type='.$col->{type}	.', col_code='.$col->{code}
						.', can`t be uses in range=>'
			);
		}		
		
		next unless defined $wcol;
		
		## store col ##
		$wcol.=' ) ';
		if($col->{for_having} && $h{group}){
			push @hcols,$wcol;
		}else{
			push @wcols,$wcol;
		}
		## /store col ##
	}
	
	

	# init limit #	
	if (0	< $r->{count})
	{## was inside the next block;
		$sql_l	= $r->{from}	> 0?($r->{from}	- 1):0;
		$sql_l .= ','.$r->{count};
	}
	if( !$h{'no_count'} && ($r->{count}>0 || $r->{select_only_count})){
		# select count #		
		if ( $h{distinct}||$r->{distinct} ) {	
			## see bug 30568 IFNULL was setted
			$sql = "SELECT COUNT(DISTINCT ".join(',',map { " IFNULL($_,'') " } @counter_cols).") FROM $sql_t $h{join}";
		}else {
			$sql = "SELECT COUNT(";
			if($h{group}){
				my $group_str = $h{group};
				my @groups;
				while (length($group_str)) {
					my $parenthesis = 0;
					my $count = 0;
					while( $group_str =~ /(.)/g) {
						$count++;
						if ( $1 eq '(' ) {
							$parenthesis++;
							next;
						}
						if ( $1 eq ')' ) {
							$parenthesis--;
							next;
						}
						if ( $1 eq ',' && !$parenthesis ) {
							$count--;
							last;
						}
					}

					my $group = substr($group_str,0,$count);
					$group =~ s/^\s?(.+)\s?$/\1/;
					push(@groups, $group);
					$group_str=substr($group_str,$count+1);
				}
				$sql.= "DISTINCT ".join(',',map { " IFNULL($_,'') " } @groups);
			}else{
				$sql.= '*';
			}
			$sql.= ") FROM $sql_t $h{join}";	
		}		
		
		#### where & having processing ###
		my $where_posted;		
		if( $h{where} ){
			$sql.=" WHERE ";
			$sql.=" $h{where} ";
			$where_posted=1;			
		}	
		if( $#wcols>=0 ){			
			$sql.=" WHERE " unless $where_posted;
			$sql.=' AND ' if $where_posted;
			$sql.= join(' AND ', @wcols );
			$where_posted=1;
		}
		if(defined $r->{where_count_sql}){
			$sql.=" WHERE " unless $where_posted;
			$sql.=' AND ' if $where_posted;
			$sql.= $r->{where_count_sql};
			$where_posted=1;
		}
		
		if( $#hcols>=0 && $h{group} ){
			$sql.=" HAVING ".join(' AND ',@hcols);
		};
		#### /where & having processing ###
		
		$r->{total_items} = $self->select_result($sql);
	}else{
		$r->{total_items} = 0;
	}	
	
	return $r->{total_items} if $r->{select_only_count};
	
	# make sql-statement #
	$sql = '';

	# distinct
	my $sql_d = ($h{distinct}||$r->{distinct}) ? 'DISTINCT' : '';
	
	$sql.=" SELECT $sql_d $sql_c FROM $sql_t $h{join}";
	
	
	#### where & having processing ###
	{
		my $where_posted;		
		if( $h{where} ){
			$sql.=" WHERE ";
			$sql.=" $h{where} ";
			$where_posted=1;			
		}	
		if( $#wcols>=0 ){			
			$sql.=" WHERE " unless $where_posted;
			$sql.=' AND ' if $where_posted;
			$sql.= join(' AND ',@wcols);
			$where_posted=1;
		}
		
		$sql.=" GROUP BY $h{group}" if $h{group};
		
		if( $#hcols>=0 && $h{group} ){
			$sql.=" HAVING ".join(' AND ',@hcols);;
		}elsif( $r->{having_sql} ){		
			$sql.=" HAVING ".$r->{having_sql};
		}				
		
		#### /where & having processing ###
	}
	
	$sql.=" ORDER BY $sql_o" if $sql_o ne '';
	$sql.=" LIMIT $sql_l" if $sql_l;								

	$sql.=" FOR UPDATE" if $h{select_for_update};

	
	# get data #
	if($r->{select_as_hashes}){
				
		$r->{data} = $self->select_rows($sql);
		my $rows;		
				
		foreach my $row( @{$r->{data}} ){			
			my $hash;			
		
			foreach my $col(@{$r->{cols}}){
				next if $col->{no_select};				
										
				$hash->{$col->{code}} = $row->[ $col->{loaded_index} ];
			}
			
			push  @$rows,$hash;
		}
											
		$r->{data} = $rows;										
	}elsif($r->{select_as_col}){
		$r->{data} = $self->select_column($sql);	
	}else{
		$r->{data} = $self->select_rows($sql);	
	}
	# calc items	
	if( !$r->{count} ){
		$r->{total_items} = $#{$r->{data}} + 1;
	}
			
	if($r->{hash_key_code}){
		# convert results to HASH #
		my $hash;		
		
		if($r->{select_as_hashes}){			
			my $code = $r->{hash_key_code};
						
			foreach my $row( @{$r->{data}} ){			
				$hash->{$row->{$code}} = $row;
			}
		}else{
			
			my $col = $r->get_col_by_code($r->{hash_key_code});
			if(!$col){
				throw HSPC::Exc::Fatal( msg => 
					qq{ERROR : no hash_key_code=$r->{hash_key_code} column in select_data()}
				);
			}
			my $key_index = $col->{index};					
			foreach	 my $res (@{ $r->{data} }){
				$hash->{$res->[$key_index]} = $res;
			}		
		}
		my $count = $#{ $r->{data} };
		$r->{data} = $hash;			
		return $count;		
	}else{
		return $#{ $r->{data} };
	}
	
}

# OPEN TRANSACTION #
# < undef=OK or 'error_string'
sub select_open{
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my $dbh = $self->db_connect;
			
	$ENV{HSPC_DB_TRANS} = {} unless ref($ENV{HSPC_DB_TRANS});
	$ENV{HSPC_DB_TRANS}->{$self->dbname}++;
	
	return undef if $ENV{HSPC_DB_TRANS}->{$self->dbname} >1;
	$ENV{HSPC_DB_ROLLBACKED} = {} unless ref($ENV{HSPC_DB_ROLLBACKED});
	$ENV{HSPC_DB_ROLLBACKED}->{$self->dbname} = 0;			
	
	$self->trace_open("OPEN TRANSACTION ($self->{dbname})");
	unless($dbh->begin_work){ 		
		return "ERROR: can't start TX: ".$dbh->errstr;
	}
	$self->trace_close();
	return undef;
}

# COMMIT TRANSACTION #
# < undef=OK or 'error_string'
sub select_commit{
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	#### check transaction counter
	
	$ENV{HSPC_DB_TRANS} = {} unless ref($ENV{HSPC_DB_TRANS});
	unless($ENV{HSPC_DB_TRANS}->{$self->dbname} >0 ){
		throw HSPC::Exc::Fatal(
			msg => 'can`t commit transaction without select_open'
		);
	}
	## decrease counter
	$ENV{HSPC_DB_TRANS}->{$self->dbname}--;	
	return unless $ENV{HSPC_DB_TRANS}->{$self->dbname}==0;	

	return $self->_db_commit();
}

sub select_rollback{		
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	#### check transaction counter
	$ENV{HSPC_DB_TRANS} = {} unless ref($ENV{HSPC_DB_TRANS});
	unless( $ENV{HSPC_DB_TRANS}->{$self->dbname}>0 ){
		throw HSPC::Exc::Fatal(
			msg => 'can`t rollback transaction without select_open'
		);
	}
	
	## decrease counter
	$ENV{HSPC_DB_TRANS}->{$self->dbname}--;	
	
	## set rollbacked flag
	$ENV{HSPC_DB_ROLLBACKED} = {} unless ref($ENV{HSPC_DB_ROLLBACKED});
	$ENV{HSPC_DB_ROLLBACKED}->{$self->dbname} = 1;
	
	## stop execution for non-zero counter
	return undef unless $ENV{HSPC_DB_TRANS}->{$self->dbname}==0;		
		
	## rollback transaction
	$self->_db_rollback();	
	
	return undef;
}

## >> class
sub clean{ 
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	
	$ENV{HSPC_DB_TRANS} = {} unless ref($ENV{HSPC_DB_TRANS});
	if( $ENV{HSPC_DB_TRANS}->{$self->dbname}){
		$self->_db_rollback();
		$ENV{HSPC_DB_TRANS}->{$self->dbname} = 0;
	};
	return undef;
}

sub _db_commit{		
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;

	return $self->_db_rollback()
		if $ENV{HSPC_DB_ROLLBACKED}
			&& $ENV{HSPC_DB_ROLLBACKED}->{$self->dbname};
	
	$self->trace_open("COMMIT TRANSACTION ($self->{dbname})");	
	my $dbh = $self->db_connect;
	unless($dbh->commit){		
		return "ERROR: can't commit transaction: ".$dbh->errstr;
	}
	$self->trace_close();	
	return undef;
}

sub _db_rollback{
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	$self->trace_open("ROLLBACK TRANSACTION ($self->{dbname})");
	$ENV{HSPC_DB_ROLLBACKED} = {} unless ref($ENV{HSPC_DB_ROLLBACKED});
	my $dbh = $self->db_connect;
	$dbh->rollback;
	$self->trace_close();	
	return undef;
}

sub select_inserted_id{
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	return $self->select_result('SELECT LAST_INSERT_ID()');
}

# => obj		# REQ_Y #
# => table		# REQ_Y #
# => gate_cols 	# REQ_Y # hash
# record key info
# => idcol	
#   or
# => id_props : ref to array with complex key properties
# /record key info
# => debug		# REQ_N #
sub delete_obj{
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;

	my %h=(@_);	
	my $obj = $h{obj};
	my $sql = qq{DELETE FROM $h{table} WHERE };		
	
	if($h{idcol}){
		my $idcol = $h{gate_cols}->{ $h{idcol} };
		$idcol=~s|^(.+\.)||; # cut table #
		$sql.= $idcol.'='.quote( $obj->{$h{idcol}} );	
	}elsif($h{id_props}){
		my $value='';
		foreach my $prop (@{ $h{id_props} }){
			$value.=q{ AND } if $value;
			$value.=  $h{gate_cols}->{$prop}.'='.quote($obj->{$prop});
		}
		$sql.=$value;
	}else{
		throw HSPC::Exc::Fatal(msg => 'ERROR: idcol=> and id_props=> arguments are both undefined');
	}
		
	
	my $dbh = $self->db_connect;							
	my $sth = $dbh->prepare($sql);						
	
	$self->trace_open($sql);
	eval{
		$sth->execute();
	};
	$self->trace_close();
	
	#### check errors
	if ($@) {
		$sth->finish();	
		throw HSPC::Exc::Mysql(msg => "$@$sql");
    }
	
	$sth->finish();
}

# => obj		# REQ_Y #
# => table		# REQ_Y # string with tables list
# => gate_cols 	# REQ_Y # hash
# record key info
# => idcol	      : ID property name
# => idcol_value  : ID property value
#   or
# => id_props : ref to array with complex key properties
# => id_props_value : ref to hash with property values {property=>value}
# /recored key info
# => where   	# REQ_N #
# => join   	# REQ_N #
# => debug		# REQ_N #
# => args       # REQ_N #

sub load_obj{
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my %h=(@_);	
	my $obj = $h{obj};
	my $sql = qq{SELECT };		
	
	# prepare cols #
	my $c;
	foreach my $prop( keys %{ $h{gate_cols} } ){
		$c.=',' if $c;
		$c.=$h{gate_cols}->{$prop}.' AS '.$prop;
	}
	$sql.=$c;
	
	$sql.=qq{ FROM $h{table} $h{join} };
	
	# HANDLE IDCOL #
	if( $h{idcol} ){
		my $idcol = $h{idcol};		
		$h{where}.=' AND ' if $h{where};				
		
		$h{where}.= $h{gate_cols}->{ $idcol  }.'=';
		$h{where}.= quote( $h{idcol_value}||$obj->{$idcol} );
	}elsif( $h{id_props} ){		
		my $w='';
		foreach my $prop (@{ $h{id_props} }){
			$w.=  ' AND ' if $w;
			$w.=  $h{gate_cols}->{$prop}.'=';
			
			if( $h{id_props_value} && defined( $h{id_props_value}->{$prop} ) ){				
				$w.=  quote( $h{id_props_value}->{$prop} );
			}else{
				$w.=  quote( $obj->{$prop} );
			}
		}
		$h{where}.=' AND ' if $h{where};				
		$h{where}.= $w;				
	}		
	
	throw HSPC::Exc::Fatal(msg=>'ERROR: where=> undefined') unless $h{where};
	$sql.=qq{ WHERE $h{where}};
	
	
	my $dbh = $self->db_connect;
	my $sth = $dbh->prepare($sql);						
	
	# process client arguments #
	if($h{args}){
		my $args = $h{args};
		for(my $i=0; $i <= $#$args; $i++)	{ $args->[$i]='' unless (defined $args->[$i]); }	
	}
	
	$self->trace_open($sql);
	eval{
		$sth->execute( @{ $h{args} } );
	};
	$self->trace_close();
	
	#### check errors
	if ($@) {
		$sth->finish();	
		throw HSPC::Exc::Mysql(msg=>"$@$sql");
    }
		
	my $h = $sth->fetchrow_hashref;
	$sth->finish();
	return 0 unless $h;
	
	map{ $obj->{$_} = $h->{$_} } keys %$h;
	
	return 1;	
}

# => obj
# => table
# => cols : array
# => gate_cols : hash
# record key info
# => idcol	
#   or
# => id_props : ref to array with complex key properties
# /recored key info
# => debug
sub update_obj {
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my %h=(@_);
	my $dbh = $self->db_connect;		
	
	my $sql = qq{UPDATE $h{table} SET };
	my $index=0;
	my $cols = $h{cols};
	my $obj = $h{obj};
	my $value;
	my @args;
			
	foreach my $col ( @$cols ){				
		$sql.=',' if $index>0;		
		
		#$value = $h{obj}->{$col}||'';
		$value = (defined $h{obj}->{$col})?$h{obj}->{$col}:'';		
		
		if($h{gate_cols}){
			$col = $h{gate_cols}->{$col}||$col;			
			$col=~s|^(.+\.)||;
		}
		$sql.="$col=?";
		push @args,$value;
		$index++;
	}			
	
	# gate to real idcol #
	if($h{idcol}){
		$h{idcol}  =  $h{gate_cols}->{ $h{idcol}  } if $h{idcol} && $h{gate_cols} && $h{gate_cols}->{ $h{idcol}  };
		$h{idcol}=~s|^(.+\.)||; # cut table #	
		$sql.=qq{ WHERE $h{idcol} = '$h{obj}->{$h{idcol}}'};
	}elsif($h{id_props}){
		$value='';
		foreach my $prop (@{ $h{id_props} }){
			$value.=q{ AND } if $value;
			if($h{gate_cols}){
				$value.=  $h{gate_cols}->{$prop}.'=?';
				push @args,$obj->{$prop};
			}else{
				$value.=  $prop.'=?';
			}
			push @args,$obj->{$prop};
		}
		$sql.=q{ WHERE }.$value;
	}else{
		throw HSPC::Exc::Fatal(msg=>"ERROR: idcol=> & id_props=> both are undefined");
	}
		
	$self->select_run($sql,@args);
	return 1;
}

# => obj 		# REQ_Y #
# => table		# REQ_Y #
# => cols 		# REQ_Y #  : array
# => gate_cols  # REQ_N #  : hash
# => idcol		# REQ_N #  : id_column name for autoincrement data
# => debug
sub insert_obj{
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my %h=(@_);
	my $obj = $h{obj};
	my $dbh = $self->db_connect;	
		
	my $index=0;
	my $cols = $h{cols};
	my $col;
	my $sql_cols;
	my $sql_values;
	my $value;
	
	foreach my $col ( @$cols ){
		$value = (defined $obj->{$col})?$obj->{$col}:'';		
		
		$sql_cols.=',' if $index>0;		
		$col = $h{gate_cols}?( $h{gate_cols}->{$col} ):$col;
		$col=~s|^(.+\.)||;
		$sql_cols.= $col;						
		
		$sql_values.=',' if $index>0;
		$sql_values .= $dbh->quote($value);
		$index++;
	}			
		
	my $sql = qq{INSERT INTO $h{table} ($sql_cols) VALUES ($sql_values)};
	
	eval{
		$dbh->do($sql);	
	};
	#### check errors
	if ($@) {
		throw HSPC::Exc::Mysql( msg => "$@$sql" );
    }
	
	$h{obj}->{ $h{idcol} } = $self->select_inserted_id if $h{idcol};
	return 1;
}

## > str
## > do not add ''
sub quote {
	my $self = ref($_[0]) ? shift : undef;
	my $str = shift;
	my $sq = shift;
	return "NULL" unless defined $str;
	$str =~ s/'/''/g;
	$str =~ s#\\#\\\\#g;
	return "$str" if ($sq);
	return "'$str'";
}

# > sql
# > params
# < hashref
sub select_hash_all {
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my $sql = shift;

	my ($row, $rec);
	my $ret = {};

	my $dbh = $self->db_connect;
	my $sth = $dbh->prepare($sql);
	
	if ( ref($_[0]) eq 'ARRAY') {
		my $set = shift;
		foreach (@$set) {
			$ret->{$_} = $rec = {};
			$sth->execute($_, @_);
			while ($row = $sth->fetchrow_arrayref) {
				$rec->{$row->[0]} = $row->[1];
			}
		}
		$sth->finish;
		return $ret;
	}
	$self->trace_open($sql);
	eval{
		$sth->execute(@_);
	};
	my $err = $@;
	$self->trace_close;
	
	#### check errors
	if ($err) {
		$sth->finish();	
		throw HSPC::Exc::Mysql( msg => "${err}${sql}" );
    }
	
	while ($row = $sth->fetchrow_arrayref) {
		$ret->{$row->[0]} = $row->[1];
	}
	$sth->finish;
	return $ret;
}


## =>> table
## < 1/0
sub is_table_exists {
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my %h = @_;

	throw HSPC::Exc::Fatal( msg => "No table param") unless $h{table};

	my $res = $self->select_rows("show tables like ?",lc($h{table}));
	if ( scalar(@{$res}) ) {
		return 1;
	} else {
		return 0;
	}
}## is_table_exists


## =>> table
## =>> column
## < 1/0
sub is_column_exists {
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my %h = @_;
	throw HSPC::Exc::Fatal( msg => "No table param") unless $h{table};
	throw HSPC::Exc::Fatal( msg => "No column param") unless $h{column};

	my $column = $self->find_column(%h);
	return $column ? 1 : 0;
}## is_column_exists


## =>> table
## =>> column
## < undef/{Field,Type,Null,Key,Default,Extra}
sub find_column {
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my %h = @_;
	throw HSPC::Exc::Fatal( msg => "No table param") unless $h{table};
	throw HSPC::Exc::Fatal( msg => "No column param") unless $h{column};

	throw HSPC::Exc::Fatal( msg => "Table ".$h{table}." not exists")
		unless $self->is_table_exists( table=>$h{table} );

	my $ar = $self->select_hashrows('desc '.lc($h{table}));

	my @res = grep {lc($_->{Field}) eq lc($h{column})} @{$ar};

	return scalar(@res) ? $res[0] : undef;
}## find_column

## Checks for key exists
## =>> table
## =>> key
## < undef/hashref with key structure
sub is_key_exists {
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my %h = @_;
	throw HSPC::Exc::Fatal( msg => "No table param") unless $h{table};
	throw HSPC::Exc::Fatal( msg => "No key param") unless $h{key};

	my $res = $self->find_keys(table => $h{table});
	my ($key) = grep { $_->{'Key_name'} eq $h{key} } @{ $res || [] };

	return $key;
}## find_keys

## Returns list of keys for table
## =>> table
## < undef/arrayref of hashrefs with key structure
sub find_keys {
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my %h = @_;
	throw HSPC::Exc::Fatal( msg => "No table param") unless $h{table};
	throw HSPC::Exc::Fatal( msg => "Table ".$h{table}." not exists")
		unless $self->is_table_exists(table => $h{table});

	my $res = $self->select_hashrows('SHOW INDEX FROM '.lc($h{table})) || [];

	return scalar(@$res) ? $res : undef;
}## find_keys

## Checks for key exists
## =>> table
## =>> key
## < undef/hashref with key structure
sub is_constraint_exists {
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my %h = @_;
	throw HSPC::Exc::Fatal(msg=>"No table param") unless length($h{table});
	throw HSPC::Exc::Fatal(msg=>"No key param") unless length($h{key});

	my $res = $self->get_create(table => $h{table});
	
	my $key = [];
	if ( $res && (ref $res eq 'HASH') ) {
		my @string = split(/\n/, $res->{'Create Table'});
			foreach my $line (@string) {
				push(@$key, [$1, $2, $3, $4] ) if ($line =~ m/^\s+CONSTRAINT\s+\`([\w\_\d]+)\`\s+FOREIGN\s+KEY\s+\(\`([\w\_\d]+)\`\)\sREFERENCES\s\`([\w\_\d]+)\`\s\(\`([\w\_\d]+)\`\)/);
			};
	}

	return (scalar grep { lc($_->[1]) eq lc($h{key}) } @$key);
}## is_constraint_exists

## Returns create table for table
## =>> table
## < undef/arrayref of hashrefs with create table structure
sub get_create {
	my $self = ref($_[0]) ? shift : new HSPC::WebDB;
	my %h = @_;
	throw HSPC::Exc::Fatal(msg=>"No table param") unless length($h{table});
	throw HSPC::Exc::Fatal(msg=>"Table ".$h{table}." not exists")
		unless $self->is_table_exists(table => $h{table});

	my $res = $self->select_hashrows('SHOW CREATE TABLE '.lc($h{table})) || [];

  my ($table) = grep { lc($_->{'Table'}) eq lc($h{table}) } @$res;
  
	return (ref $table eq 'HASH') ? $table : undef;
}## get_create

1;

__END__

# ENV VARS #

	HSPC_DB_ROLLBACKED : opened transactions count for 
							each open database (hash).

